from __future__ import annotations

import logging
import time
from datetime import timedelta
from unittest.mock import patch

import pytest

import dramatiq
from dramatiq import Message, Middleware
from dramatiq.errors import ActorNotFound, RateLimitExceeded
from dramatiq.middleware import CurrentMessage, SkipMessage, TimeLimitExceeded

from .common import skip_on_pypy, worker


def test_actors_can_be_defined(stub_broker):
    # Given that I've decorated a function with @actor
    @dramatiq.actor
    def add(x, y):
        return x + y

    # I expect that function to become an instance of Actor
    assert isinstance(add, dramatiq.Actor)


def test_actors_get_valid_logger(stub_broker):
    # Given a function with a __module__ attribute
    def add(x, y):
        return x + y

    # When I apply the @actor decorator
    decorated_add = dramatiq.actor(add)

    # I expect its logger to be named after the function's module and actor name
    assert decorated_add.logger.name == f"{add.__module__}.{add.__name__}"


def test_actors_without_module_get_valid_logger(stub_broker):
    # Given a function without a __module__ attribute
    # (e.g., a function defined in an interactive Python shell)
    def add(x, y):
        return x + y

    del add.__module__

    # When I apply the @actor decorator
    decorated_add = dramatiq.actor(add)

    # I expect its logger to be named after '_' and the actor name
    assert decorated_add.logger.name == f"_.{add.__name__}"


def test_actors_can_be_declared_with_actor_class(stub_broker):
    # Given that I have a non-standard Actor class
    class ActorChild(dramatiq.Actor):
        pass

    # When I define an actor with that class
    @dramatiq.actor(actor_class=ActorChild)
    def add(x, y):
        return x + y

    # Then that actor should be an instance of ActorChild
    assert isinstance(add, ActorChild)


def test_actors_can_be_assigned_predefined_options(stub_broker):
    # Given that I have a stub broker with the retries middleware
    # If I define an actor with a max_retries number
    @dramatiq.actor(max_retries=32)
    def add(x, y):
        return x + y

    # I expect the option to persist
    assert add.options["max_retries"] == 32


def test_actors_cannot_be_assigned_arbitrary_options(stub_broker):
    # Given that I have a stub broker
    # If I define an actor with a nonexistent option
    # I expect it to raise a ValueError
    with pytest.raises(ValueError):

        @dramatiq.actor(invalid_option=32)
        def add(x, y):
            return x + y


def test_actors_can_be_named(stub_broker):
    # Given that I've decorated a function with @actor and named it explicitly
    @dramatiq.actor(actor_name="foo")
    def add(x, y):
        return x + y

    # I expect the returned function to have that name
    assert add.actor_name == "foo"


def test_actors_can_be_assigned_custom_queues(stub_broker):
    # Given that I've decorated a function with @actor and given it an explicit queue
    @dramatiq.actor(queue_name="foo")
    def foo():
        pass

    # I expect the returned function to use that queue
    assert foo.queue_name == "foo"


def test_actors_fail_given_invalid_queue_names(stub_broker):
    # If I define an actor with an invalid queue name
    # I expect a ValueError to be raised
    with pytest.raises(ValueError):

        @dramatiq.actor(queue_name="$2@!@#")
        def foo():
            pass


def test_actors_can_be_called(stub_broker):
    # Given that I have an actor
    @dramatiq.actor
    def add(x, y):
        return x + y

    # If I call it directly,
    # I expect it to run synchronously
    assert add(1, 2) == 3


def test_actors_can_be_sent_messages(stub_broker):
    # Given that I have an actor
    @dramatiq.actor
    def add(x, y):
        return x + y

    # If I send it a message,
    # I expect it to enqueue a message
    enqueued_message = add.send(1, 2)
    enqueued_message_data = stub_broker.queues["default"].get(timeout=1)
    assert enqueued_message == Message.decode(enqueued_message_data)


def test_actors_can_perform_work(stub_broker, stub_worker):
    # Given that I have a database
    database = {}

    # And an actor that can write data to that database
    @dramatiq.actor
    def put(key, value):
        database[key] = value

    # If I send that actor many async messages
    for i in range(100):
        assert put.send("key-%s" % i, i)

    # Then join on the queue
    stub_broker.join(put.queue_name)
    stub_worker.join()

    # I expect the database to be populated
    assert len(database) == 100


def test_actors_can_perform_work_with_kwargs(stub_broker, stub_worker):
    # Given that I have a database
    results = []

    # And an actor
    @dramatiq.actor
    def add(x, y):
        results.append(x + y)

    # If I send it a message with kwargs
    add.send(x=1, y=2)

    # Then join on the queue
    stub_broker.join(add.queue_name)
    stub_worker.join()

    # I expect the database to be populated
    assert results == [3]


@skip_on_pypy
def test_actors_can_be_assigned_time_limits(stub_broker, stub_worker):
    # Given that I have a database
    attempts, successes = [], []

    # And an actor with a time limit
    @dramatiq.actor(max_retries=0, time_limit=1000)
    def do_work():
        attempts.append(1)
        time.sleep(3)
        successes.append(1)

    # When I send it a message
    do_work.send()

    # And join on the queue
    with pytest.raises(TimeLimitExceeded):  # expect TimeLimitExceeded exception
        stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # Then I expect it to fail
    assert sum(attempts) == 1
    assert sum(successes) == 0


@skip_on_pypy
def test_actor_messages_can_be_assigned_time_limits(stub_broker, stub_worker):
    # Given that I have a database
    attempts, successes = [], []

    # And an actor without an explicit time limit
    @dramatiq.actor(max_retries=0)
    def do_work():
        attempts.append(1)
        time.sleep(2)
        successes.append(1)

    # If I send it a message with a custom time limit
    do_work.send_with_options(time_limit=1000)

    # Then join on the queue
    with pytest.raises(TimeLimitExceeded):  # expect TimeLimitExceeded exception
        stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # I expect it to fail
    assert sum(attempts) == 1
    assert sum(successes) == 0


def test_actors_can_be_assigned_message_age_limits(stub_broker):
    # Given that I have a database
    runs = []

    # And an actor whose messages have an age limit
    @dramatiq.actor(max_age=100)
    def do_work():
        runs.append(1)

    # When I send it a message
    do_work.send()

    # And wait for its age limit to pass
    time.sleep(0.1)

    # Then join on its queue
    with worker(stub_broker, worker_timeout=100) as stub_worker:
        with pytest.raises(SkipMessage):  # expect SkipMessage exception
            stub_broker.join(do_work.queue_name)
        stub_worker.join()

        # I expect the message to have been skipped
        assert sum(runs) == 0


def test_actor_messages_can_be_assigned_message_age_limits(stub_broker):
    # Given that I have a database
    runs = []

    # And an actor without an explicit age limit
    @dramatiq.actor()
    def do_work():
        runs.append(1)

    # When I send it a message with a custom age limit
    do_work.send_with_options(max_age=100)

    # And wait for its age limit to pass
    time.sleep(0.1)

    # Then join on its queue
    with worker(stub_broker, worker_timeout=100) as stub_worker:
        with pytest.raises(SkipMessage):  # expect SkipMessage exception
            stub_broker.join(do_work.queue_name)
        stub_worker.join()

        # I expect the message to have been skipped
        assert sum(runs) == 0


def test_actors_can_delay_messages_independent_of_each_other(stub_broker, stub_worker):
    # Given that I have a database
    results = []

    # And an actor that appends a number to the database
    @dramatiq.actor
    def append(x):
        results.append(x)

    # If I send it a delayed message
    append.send_with_options(args=(1,), delay=1500)

    # And then another delayed message with a smaller delay and using a timedelta
    append.send_with_options(args=(2,), delay=timedelta(seconds=1))

    # Then join on the queue
    stub_broker.join(append.queue_name)
    stub_worker.join()

    # I expect the latter message to have been run first
    assert results == [2, 1]


def test_messages_belonging_to_missing_actors_are_rejected(stub_broker, stub_worker):
    # Given that I have a broker without actors
    # If I send it a message
    message = Message(
        queue_name="some-queue",
        actor_name="some-actor",
        args=(),
        kwargs={},
        options={},
    )
    stub_broker.declare_queue("some-queue")
    stub_broker.enqueue(message)

    # Then join on the queue
    with pytest.raises(ActorNotFound, match=r"^some-actor$"):
        stub_broker.join("some-queue", fail_fast=True)
    stub_worker.join()

    # I expect the message to end up on the dead letter queue
    assert stub_broker.dead_letters == [message]


def test_before_and_after_signal_failures_are_ignored(stub_broker, stub_worker):
    # Given that I have a middleware that raises exceptions when it
    # tries to process messages.
    class BrokenMiddleware(Middleware):
        def before_process_message(self, broker, message):
            raise RuntimeError("before process message error")

        def after_process_message(self, broker, message, *, result=None, exception=None):
            raise RuntimeError("after process message error")

    # And a database
    database = []

    # And an actor that appends values to the database
    @dramatiq.actor
    def append(x):
        database.append(x)

    # If add that middleware to my broker
    stub_broker.add_middleware(BrokenMiddleware())

    # And send my actor a message
    append.send(1)

    # Then join on the queue
    stub_broker.join(append.queue_name)
    stub_worker.join()

    # I expect the task to complete successfully
    assert database == [1]


def test_middleware_can_decide_to_skip_messages(stub_broker, stub_worker):
    # Given a middleware that skips all messages
    skipped_messages = []

    class SkipMiddleware(Middleware):
        def before_process_message(self, broker, message):
            raise SkipMessage()

        def after_skip_message(self, broker, message):
            skipped_messages.append(1)

    stub_broker.add_middleware(SkipMiddleware())

    # And an actor that keeps track of its calls
    calls = []

    @dramatiq.actor
    def track_call():
        calls.append(1)

    # When I send that actor a message
    track_call.send()

    # And join on the broker and the worker
    stub_broker.join(track_call.queue_name)
    stub_worker.join()

    # Then I expect the call list to be empty
    assert sum(calls) == 0

    # And the skipped_messages list to contain one item
    assert sum(skipped_messages) == 1


def test_workers_can_be_paused(stub_broker, stub_worker):
    # Given a paused worker
    stub_worker.pause()

    # And an actor that keeps track of its calls
    calls = []

    @dramatiq.actor
    def track_call():
        calls.append(1)

    # When I send that actor a message
    track_call.send()

    # And wait for 100ms
    time.sleep(0.1)

    # Then no calls should be made
    assert calls == []

    # When I resume the worker and join on it
    stub_worker.resume()
    stub_broker.join(track_call.queue_name)
    stub_worker.join()

    # Then one call should be made
    assert calls == [1]


def test_actors_can_prioritize_work(stub_broker):
    with worker(stub_broker, worker_timeout=100, worker_threads=1) as stub_worker:
        # Given that I a paused worker
        stub_worker.pause()

        # And actors with different priorities
        calls = []

        @dramatiq.actor(priority=0)
        def hi():
            calls.append("hi")

        @dramatiq.actor(priority=10)
        def lo():
            calls.append("lo")

        # When I send both actors a nubmer of messages
        for _ in range(10):
            lo.send()
            hi.send()

        # Then resume the worker and join on the queue
        stub_worker.resume()
        stub_broker.join(lo.queue_name)
        stub_worker.join()

        # Then the high priority actor should run first
        assert calls == ["hi"] * 10 + ["lo"] * 10


def test_can_call_str_on_actors(stub_broker):
    # Given that I have an actor
    @dramatiq.actor
    def test():
        pass

    # When I call str on it
    # Then I should get back its representation as a string
    assert str(test) == "Actor(test)"


def test_can_call_repr_on_actors(stub_broker):
    # Given that I have an actor
    @dramatiq.actor
    def test():
        pass

    # When I call repr on it
    # Then I should get back its representation
    assert repr(test) == "Actor(%(fn)r, queue_name='default', actor_name='test')" % vars(test)


def test_workers_log_rate_limit_exceeded_errors_differently(stub_broker, stub_worker):
    # Given that I've mocked the logging class
    with patch("logging.Logger.debug") as debug_mock:
        # And I have an actor that raises RateLimitExceeded
        @dramatiq.actor(max_retries=0)
        def raise_rate_limit_exceeded():
            raise RateLimitExceeded("exceeded")

        # When I send that actor a message
        raise_rate_limit_exceeded.send()

        # And wait for the message to get processed
        with pytest.raises(RateLimitExceeded):  # expect RateLimitExceeded exception
            stub_broker.join(raise_rate_limit_exceeded.queue_name)
        stub_worker.join()

        # Then debug mock should be called with a special message
        debug_messages = [args[0] for _, args, _ in debug_mock.mock_calls]
        assert "Rate limit exceeded in message %s: %s." in debug_messages


def test_currrent_message_middleware_exposes_the_current_message(stub_broker, stub_worker):
    # Given that I have a CurrentMessage middleware
    stub_broker.add_middleware(CurrentMessage())

    # And an actor that accesses the current message
    sent_messages = []
    received_messages = []

    @dramatiq.actor
    def accessor(x):
        message_proxy = CurrentMessage.get_current_message()
        received_messages.append(message_proxy._message)

    # When I send it a couple messages
    sent_messages.append(accessor.send(1))
    sent_messages.append(accessor.send(2))

    # And wait for it to finish its work
    stub_broker.join(accessor.queue_name)

    # Then the sent messages and the received messages should be the same
    assert sorted(sent_messages) == sorted(received_messages)

    # When I try to access the current message from a non-worker thread
    # Then I should get back None
    assert CurrentMessage.get_current_message() is None


def test_decorator_raises_error_on_duplicate_name(stub_broker):
    # Given that I have an actor named 'foo'
    @dramatiq.actor(actor_name="foo")
    def f1():
        pass

    # When I try to declare another actor with that name
    with pytest.raises(ValueError) as exc_info:

        @dramatiq.actor(actor_name="foo")
        def f2():
            pass

    # Then a ValueError should be raised
    assert exc_info.type is ValueError
    assert str(exc_info.value) == "An actor named 'foo' is already registered."


def test_worker_handles_non_numeric_eta_and_logs_warning(stub_broker, stub_worker, caplog):
    # Set the log level to capture warnings
    caplog.set_level(logging.WARNING)

    # Given an actor that records when it runs
    run = []

    @dramatiq.actor
    def record():
        run.append(True)

    # If I send it a message with a non-numeric eta manually from another project or manually from rabbitmq UI (not using dramatiq)
    message = record.message_with_options(eta="not-a-number")
    stub_broker.queues[record.queue_name].put(message.encode())

    # Then join on the queue
    stub_broker.join(record.queue_name)
    stub_worker.join()

    # I expect the message to have been processed
    assert run

    # And a warning should have been logged about the invalid eta
    assert any(
        "Invalid eta value for message" in record.message for record in caplog.records if record.levelname == "WARNING"
    )


def test_worker_handles_none_eta(stub_broker, stub_worker):
    # Given an actor that records when it runs
    run = []

    @dramatiq.actor
    def record():
        run.append(True)

    # If I send it a message with a None eta
    message = record.message_with_options(eta=None)
    stub_broker.queues[record.queue_name].put(message.encode())

    # Then join on the queue
    stub_broker.join(record.queue_name)
    stub_worker.join()

    # I expect the message to have been processed
    assert run



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import time
from typing import Optional

from ..backend import Missing, ResultBackend


class StubBackend(ResultBackend):
    """An in-memory result backend.  For use in unit tests.

    Parameters:
      namespace(str): A string with which to prefix result keys.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
    """

    results: dict[str, tuple[Optional[str], Optional[float]]] = {}

    def _get(self, message_key):
        data, expiration = self.results.get(message_key, (None, None))
        if data is not None and time.monotonic() < expiration:
            return self.encoder.decode(data)
        return Missing

    def _store(self, message_key, result, ttl):
        result_data = self.encoder.encode(result)
        expiration = time.monotonic() + int(ttl / 1000)
        self.results[message_key] = (result_data, expiration)



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from .middleware import Middleware


class Callbacks(Middleware):
    """Middleware that lets you chain success and failure callbacks
    onto Actors.

    Parameters:
      on_failure(str): The name of an actor to send a message to on
        failure.
      on_success(str): The name of an actor to send a message to on
        success.
    """

    @property
    def actor_options(self):
        return {
            "on_failure",
            "on_success",
        }

    def after_process_message(self, broker, message, *, result=None, exception=None):
        actor = broker.get_actor(message.actor_name)
        if exception is None:
            target_actor_name = message.options.get("on_success") or actor.options.get("on_success")
            if target_actor_name:
                target_actor = broker.get_actor(target_actor_name)
                target_actor.send(message.asdict(), result)

        else:
            target_actor_name = message.options.get("on_failure") or actor.options.get("on_failure")
            if target_actor_name:
                target_actor = broker.get_actor(target_actor_name)
                target_actor.send(
                    message.asdict(),
                    {
                        "type": type(exception).__name__,
                        "message": str(exception),
                    },
                )



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018,2019,2020 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import time
from collections import defaultdict
from itertools import chain
from queue import Empty, Queue
from typing import Optional

from ..broker import Broker, Consumer, MessageProxy
from ..common import current_millis, dq_name, iter_queue, join_queue
from ..errors import QueueNotFound
from ..message import Message
from ..middleware import Middleware


class StubBroker(Broker):
    """A broker that can be used within unit tests.

    Parameters:
      middleware: See :class:`Broker<dramatiq.Broker>`.
      fail_fast_default: Specifies the default value for the ``fail_fast``
        argument of :meth:`join<dramatiq.brokers.stub.StubBroker.join>`.
    """

    def __init__(self, middleware: Optional[list[Middleware]] = None, *, fail_fast_default: bool = True):
        super().__init__(middleware)

        self.dead_letters_by_queue: defaultdict[str, list[MessageProxy]] = defaultdict(list)
        self.fail_fast_default: bool = fail_fast_default

    @property
    def dead_letters(self) -> list[MessageProxy]:
        """The dead-lettered messages for all defined queues."""
        return [message for messages in self.dead_letters_by_queue.values() for message in messages]

    def consume(self, queue_name: str, prefetch: int = 1, timeout: int = 100) -> Consumer:
        """Create a new consumer for a queue.

        Parameters:
          queue_name(str): The queue to consume.
          prefetch(int): The number of messages to prefetch.
          timeout(int): The idle timeout in milliseconds.

        Raises:
          QueueNotFound: If the queue hasn't been declared.

        Returns:
          Consumer: A consumer that retrieves messages from Redis.
        """
        try:
            return _StubConsumer(
                self.queues[queue_name],
                self.dead_letters_by_queue[queue_name],
                timeout,
            )
        except KeyError:
            raise QueueNotFound(queue_name) from None

    def declare_queue(self, queue_name: str) -> None:
        """Declare a queue.  Has no effect if a queue with the given
        name has already been declared.

        Parameters:
          queue_name(str): The name of the new queue.
        """
        if queue_name in self.queues:
            return

        self.emit_before("declare_queue", queue_name)
        self.queues[queue_name] = Queue()
        self.emit_after("declare_queue", queue_name)

        delayed_name = dq_name(queue_name)
        self.queues[delayed_name] = Queue()
        self.delay_queues.add(delayed_name)
        self.emit_after("declare_delay_queue", delayed_name)

    def enqueue(self, message: Message, *, delay: Optional[int] = None) -> Message:
        """Enqueue a message.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The minimum amount of time, in milliseconds, to
            delay the message by.

        Raises:
          QueueNotFound: If the queue the message is being enqueued on
            doesn't exist.
        """
        queue_name = message.queue_name
        if delay is not None:
            queue_name = dq_name(queue_name)
            message_eta = current_millis() + delay
            message = message.copy(
                queue_name=queue_name,
                options={
                    "eta": message_eta,
                },
            )

        if queue_name not in self.queues:
            raise QueueNotFound(queue_name)

        self.emit_before("enqueue", message, delay)
        self.queues[queue_name].put(message.encode())
        self.emit_after("enqueue", message, delay)
        return message

    def flush(self, queue_name: str) -> None:
        """Drop all the messages from a queue.

        Parameters:
          queue_name(str): The queue to flush.
        """
        for _ in iter_queue(self.queues[queue_name]):
            self.queues[queue_name].task_done()

    def flush_all(self) -> None:
        """Drop all messages from all declared queues."""
        for queue_name in chain(self.queues, self.delay_queues):
            self.flush(queue_name)

        self.dead_letters_by_queue.clear()

    def join(self, queue_name: str, *, timeout: Optional[int] = None, fail_fast: Optional[bool] = None) -> None:
        """Wait for all the messages on the given queue to be
        processed.  This method is only meant to be used in tests
        to wait for all the messages in a queue to be processed.

        Raises:
          QueueJoinTimeout: When the timeout elapses.
          QueueNotFound: If the given queue was never declared.

        Parameters:
          queue_name(str): The queue to wait on.
          fail_fast(bool): When this is True and any message gets
            dead-lettered during the join, then an exception will be
            raised. When False, no exception will be raised.
            Defaults to None, which means use the value of the
            ``fail_fast_default`` instance attribute
            (which defaults to True).
          timeout(Optional[int]): The max amount of time, in
            milliseconds, to wait on this queue.

        .. versionchanged:: 2.0.0
           The ``fail_fast`` parameter now defaults to ``self.fail_fast_default``
           (which defaults to True).
        """
        try:
            queues = [
                self.queues[queue_name],
                self.queues[dq_name(queue_name)],
            ]
        except KeyError:
            raise QueueNotFound(queue_name) from None

        deadline = timeout and time.monotonic() + timeout / 1000
        should_fail_fast = fail_fast if fail_fast is not None else self.fail_fast_default
        while True:
            for queue in queues:
                join_timeout = deadline and deadline - time.monotonic()
                join_queue(queue, timeout=join_timeout)

            # We cycle through $queue then $queue.DQ then $queue
            # again in case the messages that were on the DQ got
            # moved back on $queue.
            for queue in queues:
                if queue.unfinished_tasks:
                    break
            else:
                if should_fail_fast:
                    for message in self.dead_letters_by_queue[queue_name]:
                        raise (message._exception or Exception("Message failed with unknown error")) from None

                return


class _StubConsumer(Consumer):
    def __init__(self, queue, dead_letters, timeout):
        self.queue = queue
        self.dead_letters = dead_letters
        self.timeout = timeout

    def ack(self, message):
        self.queue.task_done()

    def nack(self, message):
        self.queue.task_done()
        self.dead_letters.append(message)

    def __next__(self):
        try:
            data = self.queue.get(timeout=self.timeout / 1000)
            message = Message.decode(data)
            return _StubMessageProxy(message)
        except Empty:
            return None


class _StubMessageProxy(MessageProxy):
    def clear_exception(self) -> None:
        """Let the GC handle the cycle once the message is no longer
        in use.  This lets us keep showing full stack traces in
        failing tests.  See comment in `Worker' for details.
        """



from __future__ import annotations

import time
from unittest import mock

import pytest
import redis

import dramatiq
from dramatiq import Message, QueueJoinTimeout
from dramatiq.brokers.redis import MAINTENANCE_SCALE, RedisBroker
from dramatiq.common import current_millis, dq_name, xq_name
from dramatiq.errors import ConnectionError

from .common import worker

LUA_MAX_UNPACK_SIZE = 7999


def test_redis_actors_can_be_sent_messages(redis_broker, redis_worker):
    # Given that I have a database
    database = {}

    # And an actor that can write data to that database
    @dramatiq.actor()
    def put(key, value):
        database[key] = value

    # If I send that actor many async messages
    for i in range(100):
        assert put.send("key-%s" % i, i)

    # And I give the workers time to process the messages
    redis_broker.join(put.queue_name)
    redis_worker.join()

    # I expect the database to be populated
    assert len(database) == 100


def test_redis_actors_retry_with_backoff_on_failure(redis_broker, redis_worker):
    # Given that I have a database
    failure_time, success_time = None, None

    # And an actor that fails the first time it's called
    @dramatiq.actor(min_backoff=1000, max_backoff=5000, max_retries=1)
    def do_work():
        nonlocal failure_time, success_time
        if not failure_time:
            failure_time = current_millis()
            raise RuntimeError("First failure.")
        else:
            success_time = current_millis()

    # If I send it a message
    do_work.send()

    # Then join on the queue
    redis_broker.join(do_work.queue_name)
    redis_worker.join()

    # I expect backoff time to have passed between success and failure
    # + the worker idle timeout as padding in case the worker is long polling
    assert 500 <= success_time - failure_time <= 2500 + redis_worker.worker_timeout


def test_redis_actors_can_retry_multiple_times(redis_broker, redis_worker):
    # Given that I have a database
    attempts = []

    # And an actor that fails 3 times then succeeds
    @dramatiq.actor(min_backoff=1000, max_backoff=1000)
    def do_work():
        attempts.append(1)
        if sum(attempts) < 4:
            raise RuntimeError("Failure #%s" % sum(attempts))

    # If I send it a message
    do_work.send()

    # Then join on the queue
    redis_broker.join(do_work.queue_name)
    redis_worker.join()

    # I expect it to have been attempted 4 times
    assert sum(attempts) == 4


def test_redis_actors_can_have_their_messages_delayed(redis_broker, redis_worker):
    # Given that I have a database
    start_time, run_time = current_millis(), None

    # And an actor that records the time it ran
    @dramatiq.actor()
    def record():
        nonlocal run_time
        run_time = current_millis()

    # If I send it a delayed message
    record.send_with_options(delay=1000)

    # Then join on the queue
    redis_broker.join(record.queue_name)
    redis_worker.join()

    # I expect that message to have been processed at least delayed milliseconds later
    assert run_time - start_time >= 1000


def test_redis_actors_can_delay_messages_independent_of_each_other(redis_broker):
    # Given that I have a database
    results = []

    # And an actor that appends a number to the database
    @dramatiq.actor()
    def append(x):
        results.append(x)

    # When I pause the worker
    with worker(redis_broker, worker_timeout=100, worker_threads=1) as redis_worker:
        redis_worker.pause()

        # And I send it a delayed message
        append.send_with_options(args=(1,), delay=2000)

        # And then another delayed message with a smaller delay
        append.send_with_options(args=(2,), delay=1000)

        # Then resume the worker and join on the queue
        redis_worker.resume()
        redis_broker.join(append.queue_name)
        redis_worker.join()

        # I expect the latter message to have been run first
        assert results == [2, 1]


def test_redis_unacked_messages_can_be_requeued(redis_broker):
    # Given that I have a Redis broker
    queue_name = "some-queue"
    redis_broker.declare_queue(queue_name)

    num_messages = LUA_MAX_UNPACK_SIZE * 2
    # The lua max stack size is 8000, so try to work with double that
    message_ids = [("message-%s" % i).encode() for i in range(num_messages)]

    # If I enqueue many messages
    for message_id in message_ids:
        redis_broker.do_enqueue(queue_name, message_id, b"message-data")

    # And then fetch them
    for _ in message_ids:
        redis_broker.do_fetch(queue_name, 1)

    # Then both must be in the acks set
    ack_group = "dramatiq:__acks__.%s.%s" % (redis_broker.broker_id, queue_name)
    unacked = redis_broker.client.smembers(ack_group)
    assert sorted(unacked) == sorted(message_ids)

    # When I close that broker and open another and dispatch a command
    redis_broker.broker_id = "some-other-id"
    redis_broker.heartbeat_timeout = 0
    redis_broker.maintenance_chance = MAINTENANCE_SCALE
    redis_broker.do_qsize(queue_name)

    # Then all messages should be requeued
    ack_group = "dramatiq:__acks__.%s.%s" % (redis_broker.broker_id, queue_name)
    unacked = redis_broker.client.smembers(ack_group)
    assert not unacked

    queued = redis_broker.client.lrange("dramatiq:%s" % queue_name, 0, num_messages)
    assert set(message_ids) == set(queued)


def test_redis_messages_can_be_dead_lettered(redis_broker, redis_worker):
    # Given that I have an actor that always fails
    @dramatiq.actor(max_retries=0)
    def do_work():
        raise RuntimeError("failed")

    # If I send it a message
    do_work.send()

    # And then join on its queue
    redis_broker.join(do_work.queue_name)
    redis_worker.join()

    # I expect it to end up in the dead letter queue
    dead_queue_name = "dramatiq:%s" % xq_name(do_work.queue_name)
    dead_ids = redis_broker.client.zrangebyscore(dead_queue_name, 0, "+inf")
    assert dead_ids


def test_redis_dead_lettered_messages_are_cleaned_up(redis_broker, redis_worker):
    # Given that I have an actor that always fails
    @dramatiq.actor(max_retries=0)
    def do_work():
        raise RuntimeError("failed")

    # When I send it many messages
    for _ in range(LUA_MAX_UNPACK_SIZE * 2):
        do_work.send()

    # And then join on its queue
    redis_broker.join(do_work.queue_name)
    redis_worker.join()

    # And trigger maintenance
    redis_broker.dead_message_ttl = 0
    redis_broker.maintenance_chance = MAINTENANCE_SCALE
    redis_broker.do_qsize(do_work.queue_name)

    # Then all the messages should be removed from the DLQ.
    dead_queue_name = "dramatiq:%s" % xq_name(do_work.queue_name)
    dead_ids = redis_broker.client.zrangebyscore(dead_queue_name, 0, "+inf")
    assert not dead_ids


def test_redis_messages_belonging_to_missing_actors_are_rejected(redis_broker, redis_worker):
    # Given that I have a broker without actors
    # If I send it a message
    message = Message(
        queue_name="some-queue",
        actor_name="some-actor",
        args=(),
        kwargs={},
        options={},
    )
    redis_broker.declare_queue("some-queue")
    message = redis_broker.enqueue(message)

    # Then join on the queue
    redis_broker.join("some-queue")
    redis_worker.join()

    # I expect the message to end up on the dead letter queue
    dead_queue_name = "dramatiq:%s" % xq_name("some-queue")
    dead_ids = redis_broker.client.zrangebyscore(dead_queue_name, 0, "+inf")
    assert message.options["redis_message_id"].encode("utf-8") in dead_ids


def test_redis_requeues_unhandled_messages_on_shutdown(redis_broker):
    # Given that I have an actor that takes its time
    @dramatiq.actor
    def do_work():
        time.sleep(1)

    # If I send it two messages
    message_1 = do_work.send()
    message_2 = do_work.send()

    # Then start a worker and subsequently shut it down
    with worker(redis_broker, worker_threads=1):
        time.sleep(0.25)

    # I expect it to have processed one of the messages and re-enqueued the other
    messages = redis_broker.client.lrange("dramatiq:%s" % do_work.queue_name, 0, 10)
    if message_1.options["redis_message_id"].encode("utf-8") not in messages:
        assert message_2.options["redis_message_id"].encode("utf-8") in messages

    else:
        assert message_1.options["redis_message_id"].encode("utf-8") in messages


def test_redis_requeues_unhandled_delay_messages_on_shutdown(redis_broker):
    # Given that I have an actor that takes its time
    @dramatiq.actor
    def do_work():
        pass

    # If I send it a delayed message
    message = do_work.send_with_options(delay=10000)

    # Then start a worker and subsequently shut it down
    with worker(redis_broker, worker_threads=1):
        pass

    # I expect it to have re-enqueued the message
    messages = redis_broker.client.lrange("dramatiq:%s" % dq_name(do_work.queue_name), 0, 10)
    assert message.options["redis_message_id"].encode("utf-8") in messages


def test_redis_broker_can_join_with_timeout(redis_broker, redis_worker):
    # Given that I have an actor that takes a long time to run
    @dramatiq.actor
    def do_work():
        time.sleep(1)

    # When I send that actor a message
    do_work.send()

    # And join on its queue with a timeout
    # Then I expect a QueueJoinTimeout to be raised
    with pytest.raises(QueueJoinTimeout):
        redis_broker.join(do_work.queue_name, timeout=500)


def test_redis_broker_can_flush_queues(redis_broker):
    # Given that I have an actor
    @dramatiq.actor
    def do_work():
        pass

    # When I send that actor a message
    do_work.send()

    # And then tell the broker to flush all queues
    redis_broker.flush_all()

    # And then join on the actors's queue
    # Then it should join immediately
    assert redis_broker.join(do_work.queue_name, timeout=200) is None


def test_redis_broker_can_connect_via_url():
    # Given that I have a connection string
    # When I pass that to RedisBroker
    broker = RedisBroker(url="redis://127.0.0.1")

    # Then I should get back a valid connection
    assert broker.client.ping()


def test_redis_broker_can_connect_via_client():
    # Given that I already have redis client in hand
    # When I pass it to RedisBroker
    client = redis.Redis()
    broker = RedisBroker(client=client)

    # Then I should get back a valid connection
    assert broker.client.ping()
    assert broker.client is client


def test_redis_broker_raises_attribute_error_when_given_an_invalid_attribute(redis_broker):
    # Given that I have a Redis broker
    # When I try to access an attribute that doesn't exist
    # Then I should get back an attribute error
    with pytest.raises(AttributeError):
        redis_broker.idontexist


def test_redis_consumer_ack_can_retry_on_connection_error(redis_broker, redis_worker):
    # Given that I have an actor
    @dramatiq.actor
    def do_work():
        pass

    # If I send that actor an async message
    message = do_work.send()
    consumer = redis_worker.consumers[do_work.queue_name].consumer

    do_ack = redis_broker.do_ack

    with mock.patch.object(consumer.broker, "do_ack") as ack_mock:

        def side_effect(queue, msg_id):
            if ack_mock.call_count == 1:
                # On the first try, I expect there to be an outstanding message
                assert consumer.outstanding_message_count == 1
            else:
                # On each subsequent try, there should be 0 messages
                assert consumer.outstanding_message_count == 0

            result = do_ack(queue, msg_id)

            if ack_mock.call_count < 2:
                # Trigger a retry by raising a ConnectionError
                raise ConnectionError(message)

            return result

        ack_mock.side_effect = side_effect

        # And I give the workers time to process the messages
        redis_broker.join(do_work.queue_name)
        redis_worker.join()

        # I expect `do_ack` to have been called at least 2 times
        assert ack_mock.call_count >= 2
        # And I expect there to be no outstanding messages
        assert consumer.outstanding_message_count == 0


def test_redis_consumer_nack_can_retry_on_connection_error(redis_broker, redis_worker):
    # Given that I have an actor
    @dramatiq.actor(max_retries=0)
    def do_work():
        raise RuntimeError

    # If I send that actor an async message
    message = do_work.send()
    consumer = redis_worker.consumers[do_work.queue_name].consumer

    do_nack = redis_broker.do_nack

    with mock.patch.object(consumer.broker, "do_nack") as nack_mock:

        def side_effect(queue, msg_id):
            if nack_mock.call_count == 1:
                # On the first try, I expect there to be an outstanding message
                assert consumer.outstanding_message_count == 1
            else:
                # On each subsequent try, there should be 0 messages
                assert consumer.outstanding_message_count == 0

            result = do_nack(queue, msg_id)

            if nack_mock.call_count < 2:
                # Trigger a retry by raising a ConnectionError
                raise ConnectionError(message)

            return result

        nack_mock.side_effect = side_effect

        # And I give the workers time to process the messages
        redis_broker.join(do_work.queue_name)
        redis_worker.join()

        # I expect `do_nack` to have been called at least 2 times
        assert nack_mock.call_count >= 2
        # And I expect there to be no outstanding messages
        assert consumer.outstanding_message_count == 0


def test_redis_consumer_nack_does_not_raise_on_missing_id(redis_worker):
    """
    Reproduces issue 320, exception when the message id is not found.

    https://github.com/Bogdanp/dramatiq/issues/320

    Before the change would raise:
    redis.exceptions.ResponseError: Error running script (call to f_e9668bc413bd4a2d63c8108b124f5b7df0d01263):
    @user_script:218: @user_script: 218: Lua redis() command arguments must be strings or integers
    """

    # Given that I have an actor
    @dramatiq.actor(max_retries=0)
    def do_work():
        raise RuntimeError

    consumer = redis_worker.consumers[do_work.queue_name].consumer

    # If I send a bogus message to nack, I expect no exception to be raised.
    message = Message(
        queue_name=do_work.queue_name,
        actor_name=do_work.actor_name,
        args=(),
        kwargs={},
        options={"redis_message_id": "XXXXXXXXX"},
    )  # Bogus ID
    consumer.nack(message)


def test_redis_join_race_condition(redis_broker):
    """
    test for issue https://github.com/Bogdanp/dramatiq/issues/480

    this test verifies that do_qsize counts all messages in the queue and delay queue at the same
    time, removing the race condition mentioned in the issue
    """
    called = False
    size = []

    @dramatiq.actor
    def go():
        go_again.send_with_options(delay=1000)
        time.sleep(0.25)
        # go ack + go msg + go_again dq.msg + go_again db.ack
        size.append(redis_broker.do_qsize("default"))
        size.append(redis_broker.do_qsize(dq_name("default")))  # does the same

    @dramatiq.actor
    def go_again():
        nonlocal called
        called = True
        size.append(redis_broker.do_qsize("default"))  # go_again msg + go_again ack
        size.append(redis_broker.do_qsize(dq_name("default")))  # does the same

    with worker(redis_broker, worker_timeout=50, worker_threads=1) as redis_worker:
        go.send()
        redis_broker.join("default")
        redis_worker.join()

    assert called
    assert size == [4, 4, 2, 2]



# This file is a part of Dramatiq.
#
# Copyright (C) 2020 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from .errors import ResultFailure

# We need to deal with backwards-compatibility here in case the user
# is migrating between dramatiq versions so the canary is used to
# detect exception results.
_CANARY = "dramatiq.results.Result"


def wrap_result(res):
    # This is a no-op for forwards-compatibility.  If the user deploys
    # a new version of dramatiq and then is forced to roll back, then
    # this will minimize the fallout.
    return res


def wrap_exception(e):
    return {
        "__t": _CANARY,
        "exn": {
            "type": type(e).__name__,
            "msg": str(e),
        },
    }


def unwrap_result(res):
    if isinstance(res, dict) and res.get("__t") == _CANARY:
        message = "actor raised %s: %s" % (res["exn"]["type"], res["exn"]["msg"])
        raise ResultFailure(message, res["exn"]["type"], res["exn"]["msg"])
    return res



from __future__ import annotations

import logging
import os
import signal

import watchdog.events
import watchdog.observers.polling

try:
    import watchdog_gevent

    EVENTED_OBSERVER = watchdog_gevent.Observer
except ImportError:
    EVENTED_OBSERVER = watchdog.observers.Observer


def setup_file_watcher(path, use_polling=False, include_patterns=None, exclude_patterns=None):
    """Sets up a background thread that watches for source changes and
    automatically sends SIGHUP to the current process whenever a file
    changes.
    """
    if use_polling:
        observer_class = watchdog.observers.polling.PollingObserver
    else:
        observer_class = EVENTED_OBSERVER

    if include_patterns is None:
        include_patterns = ["*.py"]

    file_event_handler = _SourceChangesHandler(patterns=include_patterns, ignore_patterns=exclude_patterns)
    file_watcher = observer_class()
    file_watcher.schedule(file_event_handler, path, recursive=True)
    file_watcher.start()
    return file_watcher


class _SourceChangesHandler(watchdog.events.PatternMatchingEventHandler):  # pragma: no cover
    """Handles source code change events by sending a HUP signal to
    the current process.
    """

    def on_any_event(self, event: watchdog.events.FileSystemEvent):
        # watchdog >= 2.3 emits an extra event when a file is opened that will cause unnecessary
        # restarts.  This affects any usage of watchdog where a custom event handler is used, including
        # this _SourceChangesHandler.
        #
        # For more context, see:
        #   https://github.com/gorakhargosh/watchdog/issues/949
        #   https://github.com/pallets/werkzeug/pull/2604/files
        if event.event_type == "opened":
            return

        # watchdog >= 5.0.0 added events for when a file is closed without being written.
        # Also ignore these so they don't cause unnecessay restarts.
        # See https://github.com/gorakhargosh/watchdog/pull/1059
        if event.event_type == "closed_no_write":
            return

        logger = logging.getLogger("SourceChangesHandler")
        logger.info("Detected changes to %r.", event.src_path)
        os.kill(os.getpid(), signal.SIGHUP)



from __future__ import annotations

import os
import signal
import time

from dramatiq.brokers.stub import StubBroker

from .common import skip_on_windows

broker = StubBroker()


def remove(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


@skip_on_windows
def test_cli_scrubs_stale_pid_files(start_cli):
    try:
        # Given that I have an existing file containing an old pid
        filename = "test_scrub.pid"
        with open(filename, "w") as f:
            f.write("999999")

        # When I try to start the cli and pass that file as a PID file
        proc = start_cli("tests.test_pidfile:broker", extra_args=["--pid-file", filename])

        # And I wait for it to write the pid file
        time.sleep(1)

        # Then the process should write its pid to the file
        with open(filename, "r") as f:
            pid = int(f.read())

        assert pid == proc.pid

        # When I stop the process
        proc.terminate()
        proc.wait()

        # Then the process should exit with return code 0
        assert proc.returncode == 0

        # And the file should be removed
        assert not os.path.exists(filename)
    finally:
        remove(filename)


def test_cli_aborts_when_pidfile_contains_garbage(start_cli):
    try:
        # Given that I have an existing file containing important information
        filename = "test_garbage.pid"
        with open(filename, "w") as f:
            f.write("important!")

        # When I try to start the cli and pass that file as a PID file
        proc = start_cli("tests.test_pidfile:broker", extra_args=["--pid-file", filename])
        proc.wait()

        # Then the process should exit with return code 4
        assert proc.returncode == 4
    finally:
        remove(filename)


@skip_on_windows
def test_cli_with_pidfile_can_be_reloaded(start_cli):
    try:
        # Given that I have a PID file
        filename = "test_reload.pid"

        # When I try to start the cli and pass that file as a PID file
        proc = start_cli("tests.test_pidfile:broker", extra_args=["--pid-file", filename])
        time.sleep(1)

        # And send the proc a HUP signal
        proc.send_signal(signal.SIGHUP)
        time.sleep(5)

        # And then terminate the process
        proc.terminate()
        proc.wait()

        # Then the process should exit with return code 0
        assert proc.returncode == 0
    finally:
        remove(filename)



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import threading
import warnings

from ..logging import get_logger
from .middleware import Middleware
from .threading import (
    Interrupt,
    current_platform,
    is_gevent_active,
    raise_thread_exception,
    supported_platforms,
)


class Shutdown(Interrupt):
    """Exception used to interrupt worker threads when their worker
    processes have been signaled for termination.
    """


class ShutdownNotifications(Middleware):
    """Middleware that interrupts actors whose worker process has been
    signaled for termination.
    Currently, this is only available on CPython.

    Note:
      This works by setting an async exception in the worker thread
      that runs the actor.  This means that the exception will only get
      called the next time that thread acquires the GIL.  Concretely,
      this means that this middleware can't cancel system calls.

    Parameters:
      notify_shutdown(bool): When True, the actor will be interrupted
        if the worker process was terminated.
        Defaults to False, meaning actors will not be interrupted, and allowed to finish.
    """

    def __init__(self, notify_shutdown: bool = False) -> None:
        self.logger = get_logger(__name__, type(self))
        self.notify_shutdown = notify_shutdown

        self.manager: _ShutdownManager
        if is_gevent_active():
            self.manager = _GeventShutdownManager(self.logger)
        else:
            self.manager = _CtypesShutdownManager(self.logger)

    @property
    def actor_options(self):
        return {"notify_shutdown"}

    def should_notify(self, actor, message):
        notify = message.options.get("notify_shutdown")
        if notify is None:
            notify = actor.options.get("notify_shutdown")
        if notify is None:
            notify = self.notify_shutdown
        return bool(notify)

    def after_process_boot(self, broker):
        if current_platform not in supported_platforms:  # pragma: no cover
            msg = "ShutdownNotifications cannot kill threads on your current platform (%r)."
            warnings.warn(msg % current_platform, category=RuntimeWarning, stacklevel=2)

    def before_worker_shutdown(self, broker, worker):
        self.logger.debug("Sending shutdown notification to worker threads...")
        self.manager.shutdown()

    def before_process_message(self, broker, message):
        actor = broker.get_actor(message.actor_name)

        if self.should_notify(actor, message):
            self.manager.add_notification()

    def after_process_message(self, broker, message, *, result=None, exception=None):
        self.manager.remove_notification()

    after_skip_message = after_process_message


class _ShutdownManager:

    def __init__(self, logger=None) -> None:  # pragma: no cover
        raise NotImplementedError

    def add_notification(self) -> None:  # pragma: no cover
        raise NotImplementedError

    def remove_notification(self) -> None:  # pragma: no cover
        raise NotImplementedError

    def shutdown(self) -> None:  # pragma: no cover
        raise NotImplementedError


class _CtypesShutdownManager(_ShutdownManager):

    def __init__(self, logger=None):
        self.logger = logger or get_logger(__name__, type(self))
        self.notifications = set()

    def add_notification(self):
        self.notifications.add(threading.get_ident())

    def remove_notification(self):
        self.notifications.discard(threading.get_ident())

    def shutdown(self):
        for thread_id in self.notifications:
            self.logger.info(
                "Worker shutdown notification. Raising exception in worker thread %r.",
                thread_id,
            )
            raise_thread_exception(thread_id, Shutdown)


if is_gevent_active():
    from gevent import getcurrent

    class _GeventShutdownManager(_ShutdownManager):

        def __init__(self, logger=None):
            self.logger = logger or get_logger(__name__, type(self))
            self.notification_greenlets = set()

        def add_notification(self):
            current_greenlet = getcurrent()
            # Get and store the threading ident rather than using the greenlet's
            # minimal_ident for logging consistency with the time limit middleware.
            thread_id = threading.get_ident()
            self.notification_greenlets.add((thread_id, current_greenlet))

        def remove_notification(self):
            current_greenlet = getcurrent()
            thread_id = threading.get_ident()
            self.notification_greenlets.discard((thread_id, current_greenlet))

        def shutdown(self):
            for thread_id, greenlet in self.notification_greenlets:
                self.logger.info(
                    "Worker shutdown notification. Raising exception in worker thread %r.",
                    thread_id,
                )
                greenlet.kill(Shutdown, block=False)



from __future__ import annotations

import pytest

from dramatiq.common import dq_name, q_name, xq_name


@pytest.mark.parametrize(
    "given,expected",
    [
        ("default", "default"),
        ("default.DQ", "default"),
        ("default.XQ", "default"),
    ],
)
def test_q_name_returns_canonical_names(given, expected):
    assert q_name(given) == expected


@pytest.mark.parametrize(
    "given,expected",
    [
        ("default", "default.DQ"),
        ("default.DQ", "default.DQ"),
        ("default.XQ", "default.DQ"),
    ],
)
def test_dq_name_returns_canonical_delay_names(given, expected):
    assert dq_name(given) == expected


@pytest.mark.parametrize(
    "given,expected",
    [
        ("default", "default.XQ"),
        ("default.DQ", "default.XQ"),
        ("default.XQ", "default.XQ"),
    ],
)
def test_xq_name_returns_delay_names(given, expected):
    assert xq_name(given) == expected



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import threading
from os import getenv
from queue import Empty
from random import uniform
from time import time
from typing import Union

from .errors import QueueJoinTimeout


def getenv_int(name: str) -> Union[int, None]:
    """Parse an optional environment variable as an integer."""
    v = getenv(name, None)
    if v is None:
        return None
    try:
        return int(v)
    except ValueError:
        raise ValueError("invalid integer value for env var %r: %r" % (name, v)) from None


def compute_backoff(
    attempts: int, *, factor: int = 5, jitter: bool = True, max_backoff: int = 2000, max_exponent: int = 32
) -> tuple[int, int]:
    """Compute an exponential backoff value based on some number of attempts.

    Parameters:
      attempts(int): The number of attempts there have been so far.
      factor(int): The number of milliseconds to multiply each backoff by.
                   This is also effectively the minimum backoff amount.
      max_backoff(int): The max number of milliseconds to backoff by.
      max_exponent(int): The maximum backoff exponent.

    Returns:
      tuple: The new number of attempts and the backoff in milliseconds.
    """
    # Calculate base backoff from number of attempts.
    # When attempts=0, then backoff=factor, i.e. the minimum backoff.
    # each extra attempt, doubles this.
    exponent = min(attempts, max_exponent)
    backoff = factor * (2**exponent)

    if jitter:
        # Add jitter to extend backoff up to twice as much.
        backoff = int(backoff * uniform(1, 2))
        # If this exceeds max_backoff, bring backoff back to 50%-100% of max_backoff to keep some jitter.
        if backoff > max_backoff:
            backoff = int(max_backoff * uniform(0.5, 1))
    else:
        # if not jittering, simply apply max_backoff
        backoff = min(backoff, max_backoff)

    return attempts + 1, backoff


def current_millis() -> int:
    """Returns the current UNIX time in milliseconds."""
    return int(time() * 1000)


def iter_queue(queue):
    """Iterate over the messages on a queue until it's empty.  Does
    not block while waiting for messages.

    Parameters:
      queue(Queue): The queue to iterate over.

    Returns:
      generator[object]: The iterator.
    """
    while True:
        try:
            yield queue.get_nowait()
        except Empty:
            break


def join_queue(queue, timeout=None):
    """The join() method of standard queues in Python doesn't support
    timeouts.  This implements the same functionality as that method,
    with optional timeout support, using only exposed Queue interfaces.

    Raises:
      QueueJoinTimeout: When the timeout is reached.

    Parameters:
      queue(Queue)
      timeout(Optional[float])
    """
    if timeout is None:
        queue.join()
        return

    join_complete = threading.Event()

    def join_and_signal():
        queue.join()
        join_complete.set()

    join_thread = threading.Thread(target=join_and_signal, name="join_and_signal_thread")
    join_thread.daemon = True
    join_thread.start()

    # Wait for the join to complete or timeout
    if not join_complete.wait(timeout):
        raise QueueJoinTimeout("timed out after %.02f seconds" % timeout)


def join_all(joinables, timeout):
    """Wait on a list of objects that can be joined with a total
    timeout represented by ``timeout``.

    Parameters:
      joinables(object): Objects with a join method.
      timeout(int): The total timeout in milliseconds.
    """
    started, elapsed = current_millis(), 0
    for ob in joinables:
        ob.join(timeout=timeout / 1000)
        elapsed = current_millis() - started
        timeout = max(0, timeout - elapsed)


def q_name(queue_name: str) -> str:
    """Returns the canonical queue name for a given queue."""
    if queue_name.endswith(".DQ") or queue_name.endswith(".XQ"):
        return queue_name[:-3]
    return queue_name


def dq_name(queue_name: str) -> str:
    """Returns the delayed queue name for a given queue.  If the given
    queue name already belongs to a delayed queue, then it is returned
    unchanged.
    """
    if queue_name.endswith(".DQ"):
        return queue_name

    if queue_name.endswith(".XQ"):
        queue_name = queue_name[:-3]
    return queue_name + ".DQ"


def xq_name(queue_name: str) -> str:
    """Returns the dead letter queue name for a given queue.  If the
    given queue name belongs to a delayed queue, the dead letter queue
    name for the original queue is generated.
    """
    if queue_name.endswith(".XQ"):
        return queue_name

    if queue_name.endswith(".DQ"):
        queue_name = queue_name[:-3]
    return queue_name + ".XQ"



from __future__ import annotations

import time
from concurrent.futures import ThreadPoolExecutor

import pytest

from dramatiq.rate_limits import Barrier


def test_barrier(rate_limiter_backend):
    # Given that I have a barrier of two parties
    barrier = Barrier(rate_limiter_backend, "sequential-barrier", ttl=30000)
    assert barrier.create(parties=2)

    # When I try to recreate it
    # Then that should return False
    assert not barrier.create(parties=10)

    # The first party to wait should get back False
    assert not barrier.wait(block=False)

    # The second party to wait should get back True
    assert barrier.wait(block=False)


def test_barriers_can_block(rate_limiter_backend):
    # Given that I have a barrier of two parties
    barrier = Barrier(rate_limiter_backend, "sequential-barrier", ttl=30000)
    assert barrier.create(parties=2)

    # And I have a worker function that waits on the barrier and writes its timestamp
    times = []

    def worker():
        time.sleep(0.1)
        assert barrier.wait(timeout=1000)
        times.append(time.monotonic())

    try:
        # When I run those workers
        with ThreadPoolExecutor(max_workers=8) as e:
            for future in [e.submit(worker), e.submit(worker)]:
                future.result()
    except NotImplementedError:
        pytest.skip("Waiting is not supported under this backend.")

    # Then their execution times should be really close to one another
    assert abs(times[0] - times[1]) <= 0.01


def test_barriers_can_timeout(rate_limiter_backend):
    # Given that I have a barrier of two parties
    barrier = Barrier(rate_limiter_backend, "sequential-barrier", ttl=30000)
    assert barrier.create(parties=2)

    try:
        # When I wait on the barrier with a timeout
        # Then I should get False back
        assert not barrier.wait(timeout=1000)
    except NotImplementedError:
        pytest.skip("Waiting is not supported under this backend.")



from __future__ import annotations

import logging

import pytest

import dramatiq
import dramatiq.broker
from dramatiq.brokers.rabbitmq import RabbitmqBroker
from dramatiq.middleware import Middleware

from .common import RABBITMQ_CREDENTIALS, skip_on_windows


class EmptyMiddleware(Middleware):
    pass


def test_broker_uses_rabbitmq_if_not_set():
    # Given that no global broker is set
    dramatiq.broker.global_broker = None

    # If I try to get the global broker
    broker = dramatiq.get_broker()

    # I expect it to be a RabbitmqBroker instance
    assert isinstance(broker, RabbitmqBroker)


@skip_on_windows
def test_broker_middleware_can_be_added_before_other_middleware(stub_broker):
    from dramatiq.middleware import AgeLimit

    # Given that I have a custom middleware
    empty_middleware = EmptyMiddleware()

    # If I add it before the AgeLimit middleware
    stub_broker.add_middleware(empty_middleware, before=AgeLimit)

    # I expect it to be the first middleware
    assert stub_broker.middleware[0] == empty_middleware


@skip_on_windows
def test_broker_middleware_can_be_added_after_other_middleware(stub_broker):
    from dramatiq.middleware import AgeLimit

    # Given that I have a custom middleware
    empty_middleware = EmptyMiddleware()

    # If I add it after the AgeLimit middleware
    stub_broker.add_middleware(empty_middleware, after=AgeLimit)

    # I expect it to be the second middleware
    assert stub_broker.middleware[1] == empty_middleware


def test_broker_middleware_can_fail_to_be_added_before_or_after_missing_middleware(stub_broker):
    # Given that I have a custom middleware
    empty_middleware = EmptyMiddleware()

    # If I add it after a middleware that isn't registered
    # I expect a ValueError to be raised
    with pytest.raises(ValueError):
        stub_broker.add_middleware(empty_middleware, after=EmptyMiddleware)


@skip_on_windows
def test_broker_middleware_cannot_be_addwed_both_before_and_after(stub_broker):
    from dramatiq.middleware import AgeLimit

    # Given that I have a custom middleware
    empty_middleware = EmptyMiddleware()

    # If I add it with both before and after parameters
    # I expect an AssertionError to be raised
    with pytest.raises(AssertionError):
        stub_broker.add_middleware(empty_middleware, before=AgeLimit, after=AgeLimit)


def test_can_instantiate_brokers_without_middleware():
    # Given that I have an empty list of middleware
    # When I pass that to the RMQ Broker
    broker = RabbitmqBroker(middleware=[], credentials=RABBITMQ_CREDENTIALS)

    # Then I should get back a broker with no middleware
    assert not broker.middleware


def test_broker_middleware_logs_warning_when_added_twice(stub_broker, caplog):
    # Set the log level to capture warnings
    caplog.set_level(logging.WARNING)

    # Given that I have a custom middleware
    empty_middleware1 = EmptyMiddleware()
    empty_middleware2 = EmptyMiddleware()

    # When I add the first middleware
    stub_broker.add_middleware(empty_middleware1)

    # And I add another middleware of the same type
    stub_broker.add_middleware(empty_middleware2)

    # Then I expect a warning to be logged
    assert any(
        "You're adding a middleware of the same type twice" in record.message
        for record in caplog.records
        if record.levelname == "WARNING"
    )

    # And I expect both middlewares to be added
    assert empty_middleware1 in stub_broker.middleware
    assert empty_middleware2 in stub_broker.middleware



from __future__ import annotations

import os
import signal
import time
from subprocess import PIPE, STDOUT

import dramatiq
from dramatiq.brokers.redis import RedisBroker
from dramatiq.common import current_millis
from dramatiq.middleware import Middleware

from .common import skip_on_windows

fakebroker = object()


class BrokerHolder:
    fakebroker = object()


broker = RedisBroker()
loaded_at = current_millis()
fork_function_signal_mask_write_path = "/tmp/dramatiq-cli-fork-function-blocked-signals"
middleware_fork_signal_mask_write_path = "/tmp/dramatiq-cli-middleware-fork-blocked-signals"
consumer_signal_mask_write_path = "/tmp/dramatiq-cli-consumer-blocked-signals"
after_process_boot_signal_mask_write_path = "/tmp/dramatiq-cli-boot-blocked-signals"


@dramatiq.actor(broker=broker)
def write_loaded_at(filename):
    with open(filename, "w") as f:
        f.write(str(loaded_at))


@dramatiq.actor(broker=broker)
def write_masked_signals(filename):
    current_sigmask = sorted(signal.pthread_sigmask(signal.SIG_BLOCK, []))
    with open(filename, "w") as f:
        f.write(f"current_sigmask={current_sigmask}")


def fork_function_write_masked_signals():
    current_sigmask = sorted(signal.pthread_sigmask(signal.SIG_BLOCK, []))
    print("Writing to fork function path")
    with open(fork_function_signal_mask_write_path, "w") as f:
        f.write(f"current_sigmask={current_sigmask}")


class ConsumerThreadMaskWritingMiddleware(Middleware):
    def before_ack(self, _broker, _message):
        current_sigmask = sorted(signal.pthread_sigmask(signal.SIG_BLOCK, []))
        with open(consumer_signal_mask_write_path, "w") as f:
            f.write(f"current_sigmask={current_sigmask}")


consumer_mask_writing_broker = RedisBroker(middleware=[ConsumerThreadMaskWritingMiddleware()])


@dramatiq.actor(broker=consumer_mask_writing_broker)
def consumer_write_masked_signals():
    return None


def fork_middleware_write_masked_signals():
    current_sigmask = sorted(signal.pthread_sigmask(signal.SIG_BLOCK, []))
    with open(middleware_fork_signal_mask_write_path, "w") as f:
        f.write(f"current_sigmask={current_sigmask}")


class ForkMaskWritingMiddleware(Middleware):
    @property
    def forks(self):
        return [fork_middleware_write_masked_signals]


middleware_fork_mask_writing_broker = RedisBroker(middleware=[ForkMaskWritingMiddleware()])


@dramatiq.actor(broker=middleware_fork_mask_writing_broker)
def middleware_fork_actor():
    return None


class AfterProcessBootMaskWritingMiddleware(Middleware):
    def after_process_boot(self, broker):
        current_sigmask = sorted(signal.pthread_sigmask(signal.SIG_BLOCK, []))
        with open(after_process_boot_signal_mask_write_path, "w") as f:
            f.write(f"current_sigmask={current_sigmask}")


after_process_boot_mask_writing_broker = RedisBroker(middleware=[AfterProcessBootMaskWritingMiddleware()])


@dramatiq.actor(broker=after_process_boot_mask_writing_broker)
def after_process_boot_actor():
    return None


@skip_on_windows
def test_cli_fails_to_start_given_an_invalid_broker_name(start_cli):
    # Given that this module doesn't define a broker called "idontexist"
    # When I start the cli and point it at that broker
    proc = start_cli("tests.test_cli:idontexist", stdout=PIPE, stderr=STDOUT)
    proc.wait(60)

    # Then the process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"Module 'tests.test_cli' does not define a 'idontexist' variable." in proc.stdout.read()


@skip_on_windows
def test_cli_fails_to_start_given_an_invalid_broker_instance(start_cli):
    # Given that this module defines a "fakebroker" variable that's not a Broker
    # When I start the cli and point it at that broker
    proc = start_cli("tests.test_cli:fakebroker", stdout=PIPE, stderr=STDOUT)
    proc.wait(60)

    # Then the process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"'tests.test_cli:fakebroker' is not a Broker." in proc.stdout.read()


@skip_on_windows
def test_cli_fails_to_start_given_an_invalid_nested_broker_instance(start_cli):
    # Given that this module defines a "BrokerHolder.fakebroker" variable that's not a Broker
    # When I start the cli and point it at that broker
    proc = start_cli("tests.test_cli:BrokerHolder.fakebroker", stdout=PIPE, stderr=STDOUT)
    proc.wait(60)

    # Then the process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"'tests.test_cli:BrokerHolder.fakebroker' is not a Broker." in proc.stdout.read()


@skip_on_windows
def test_cli_can_be_reloaded_on_sighup(start_cli):
    # Given that I have a shared file the processes can use to communicate with
    filename = "/tmp/dramatiq-cli-loaded-at"

    # When I start my workers
    proc = start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
        ],
    )

    # And enqueue a task to write the loaded timestamp
    write_loaded_at.send(filename)
    broker.join(write_loaded_at.queue_name)

    # Then I expect a timestamp to have been written to the file
    with open(filename, "r") as f:
        timestamp_1 = int(f.read())

    # When I send a SIGHUP signal
    os.kill(proc.pid, signal.SIGHUP)

    # And wait for the workers to reload
    time.sleep(5)

    # And write another timestamp
    write_loaded_at.send(filename)
    broker.join(write_loaded_at.queue_name)

    # Then I expect another timestamp to have been written to the file
    with open(filename, "r") as f:
        timestamp_2 = int(f.read())

    # And the second time to be at least a second apart from the first
    assert timestamp_2 - timestamp_1 >= 1000


@skip_on_windows
def test_worker_threads_have_no_blocked_signals(start_cli):
    # Given that I have a shared file the processes can use to communicate with
    filename = "/tmp/dramatiq-cli-worker-blocked-signals"

    # When I start my workers
    start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
        ],
    )

    # And enqueue a task to write the blocked signals
    write_masked_signals.send(filename)
    broker.join(write_masked_signals.queue_name)

    # And then read the blocked signals to written to the file by the worker
    with open(filename, "r") as f:
        blocked_signals = f.read()

    # Then I expect that no signals are blocked
    assert blocked_signals == "current_sigmask=[]"


@skip_on_windows
def test_consumer_threads_have_no_blocked_signals(start_cli):
    # Given that I have a shared file the consumer threads can use to communicate
    filename = consumer_signal_mask_write_path

    # When I start workers on a custom broker with middleware to cause consumer threads to write masked signals
    start_cli(
        "tests.test_cli:consumer_mask_writing_broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
        ],
    )

    # And enqueue a task for a worker
    consumer_write_masked_signals.send()
    consumer_mask_writing_broker.join(consumer_write_masked_signals.queue_name)

    # And then read the blocked signals written to the file by the consumer thread
    with open(filename, "r") as f:
        blocked_signals = f.read()

    # Then I expect that no signals are blocked
    assert blocked_signals == "current_sigmask=[]"


@skip_on_windows
def test_middleware_fork_functions_have_no_blocked_signals(start_cli):
    # Given that I have a shared file that the fork process can use to communicate
    filename = middleware_fork_signal_mask_write_path

    # When I start workers on a custom broker with middleware with a fork function that writes masked signals
    start_cli(
        "tests.test_cli:middleware_fork_mask_writing_broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
        ],
    )

    # And enqueue a task for a worker, to wait for worker processes to finish setting up
    middleware_fork_actor.send()
    middleware_fork_mask_writing_broker.join(middleware_fork_actor.queue_name)

    # And wait long enough for the fork function to execute
    time.sleep(2)

    # And then read the blocked signals written to the file by the fork process
    with open(filename, "r") as f:
        blocked_signals = f.read()

    # Then I expect that no signals are blocked
    assert blocked_signals == "current_sigmask=[]"


@skip_on_windows
def test_cli_fork_functions_have_no_blocked_signals(start_cli):
    # Given that I have a shared file that the fork process can use to communicate
    filename = fork_function_signal_mask_write_path

    # When I start workers and a fork function
    start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
            "--fork",
            "tests.test_cli:fork_function_write_masked_signals",
        ],
    )

    # And enqueue a task for a worker, to wait for worker processes to finish setting up
    write_masked_signals.send("/tmp/dramatiq-dummy-not-used")
    broker.join(write_masked_signals.queue_name)

    # And wait long enough for the fork function to execute
    time.sleep(2)

    # And then read the blocked signals written to the file by the fork process
    with open(filename, "r") as f:
        blocked_signals = f.read()

    # Then I expect that no signals are blocked
    assert blocked_signals == "current_sigmask=[]"


@skip_on_windows
def test_after_process_boot_call_has_no_blocked_signals(start_cli):
    # Given that I have a shared file that the fork process can use to communicate
    filename = after_process_boot_signal_mask_write_path

    # When I start workers on a custom broker with middleware with an after_process_boot
    # method that writes masked signals
    start_cli(
        "tests.test_cli:after_process_boot_mask_writing_broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
        ],
    )

    # And enqueue a task for a worker
    after_process_boot_actor.send()
    after_process_boot_mask_writing_broker.join(after_process_boot_actor.queue_name)

    # And then read the blocked signals written to the file in the after_process_boot call
    with open(filename, "r") as f:
        blocked_signals = f.read()

    # Then I expect that no signals are blocked
    assert blocked_signals == "current_sigmask=[]"


@skip_on_windows
def test_cli_worker_fork_timeout_invalid_string(start_cli):
    proc = start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
            "--worker-fork-timeout",
            "tests",
        ],
        stdout=PIPE,
        stderr=STDOUT,
    )
    proc.wait(60)

    # The process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"worker-fork-timeout be a number." in proc.stdout.read()


@skip_on_windows
def test_cli_worker_fork_timeout_invalid_str(start_cli):
    proc = start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
            "--worker-fork-timeout",
            "tests",
        ],
        stdout=PIPE,
        stderr=STDOUT,
    )
    proc.wait(60)

    # The process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"worker-fork-timeout be a number." in proc.stdout.read()


@skip_on_windows
def test_cli_worker_fork_timeout_invalid_negative(start_cli):
    proc = start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
            "--worker-fork-timeout",
            "-5",
        ],
        stdout=PIPE,
        stderr=STDOUT,
    )
    proc.wait(60)

    # The process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"worker-fork-timeout too small (minimum recommended: 10ms)." in proc.stdout.read()


@skip_on_windows
def test_cli_worker_fork_timeout_invalid_below_10(start_cli):
    proc = start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
            "--worker-fork-timeout",
            "9",
        ],
        stdout=PIPE,
        stderr=STDOUT,
    )
    proc.wait(60)

    # The process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"worker-fork-timeout too small (minimum recommended: 10ms)." in proc.stdout.read()


@skip_on_windows
def test_cli_worker_fork_timeout_invalid_above_1_800_000(start_cli):
    proc = start_cli(
        "tests.test_cli:broker",
        extra_args=[
            "--processes",
            "1",
            "--threads",
            "1",
            "--worker-fork-timeout",
            "3_600_000",
        ],
        stdout=PIPE,
        stderr=STDOUT,
    )
    proc.wait(60)

    # The process return code should be 2
    assert proc.returncode == 2

    # And the output should contain an error
    assert b"worker-fork-timeout too large (maximum: 30 minutes)." in proc.stdout.read()



from __future__ import annotations

import logging
import os
import time
from threading import Event
from unittest.mock import Mock, patch

import pika.exceptions
import pytest

import dramatiq
from dramatiq import Message, Middleware, QueueJoinTimeout, Worker
from dramatiq.brokers.rabbitmq import (
    MAX_DECLARE_ATTEMPTS,
    RabbitmqBroker,
    _IgnoreScaryLogs,
)
from dramatiq.common import current_millis

from .common import (
    RABBITMQ_CREDENTIALS,
    RABBITMQ_PASSWORD,
    RABBITMQ_USERNAME,
    skip_unless_rabbit_mq,
)


@skip_unless_rabbit_mq
def test_rabbitmq_broker_can_be_passed_a_semicolon_separated_list_of_uris():
    # Given a string with a list of RabbitMQ connection URIs, including an invalid one
    # When I pass those URIs to RabbitMQ broker as a ;-separated string
    broker = RabbitmqBroker(
        url="amqp://127.0.0.1:55672;amqp://%s:%s@127.0.0.1" % (RABBITMQ_USERNAME, RABBITMQ_PASSWORD)
    )

    # The the broker should connect to the host that is up
    assert broker.connection


@skip_unless_rabbit_mq
def test_rabbitmq_broker_can_be_passed_a_list_of_uri_for_failover():
    # Given a string with a list of RabbitMQ connection URIs, including an invalid one
    # When I pass those URIs to RabbitMQ broker as a list
    broker = RabbitmqBroker(
        url=[
            "amqp://127.0.0.1:55672",
            "amqp://%s:%s@127.0.0.1" % (RABBITMQ_USERNAME, RABBITMQ_PASSWORD),
        ]
    )

    # The the broker should connect to the host that is up
    assert broker.connection


def test_rabbitmq_broker_raises_an_error_if_given_invalid_parameter_combinations():
    # Given that I have a RabbitmqBroker
    # When I try to give it both a connection URL and a list of connection parameters
    # Then a RuntimeError should be raised
    with pytest.raises(RuntimeError):
        RabbitmqBroker(
            url="amqp://127.0.0.1:5672",
            parameters=[dict(host="127.0.0.1", credentials=RABBITMQ_CREDENTIALS)],
        )

    # When I try to give it both a connection URL and pika connection parameters
    # Then a RuntimeError should be raised
    with pytest.raises(RuntimeError):
        RabbitmqBroker(host="127.0.0.1", url="amqp://127.0.0.1:5672")

    # When I try to give it both a list of parameters and individual flags
    # Then a RuntimeError should be raised
    with pytest.raises(RuntimeError):
        RabbitmqBroker(host="127.0.0.1", parameters=[dict(host="127.0.0.1")])


@skip_unless_rabbit_mq
def test_rabbitmq_broker_can_be_passed_a_list_of_parameters_for_failover():
    # Given a list of pika connection parameters including an invalid one
    parameters = [
        dict(host="127.0.0.1", port=55672),  # this will fail
        dict(host="127.0.0.1", credentials=RABBITMQ_CREDENTIALS),
    ]

    # When I pass those parameters to RabbitmqBroker
    broker = RabbitmqBroker(parameters=parameters)

    # Then I should still get a connection to the host that is up
    assert broker.connection


def test_rabbitmq_actors_can_be_sent_messages(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    database = {}

    # And an actor that can write data to that database
    @dramatiq.actor
    def put(key, value):
        database[key] = value

    # If I send that actor many async messages
    for i in range(100):
        assert put.send("key-%d" % i, i)

    # And I give the workers time to process the messages
    rabbitmq_broker.join(put.queue_name)
    rabbitmq_worker.join()

    # I expect the database to be populated
    assert len(database) == 100


def test_rabbitmq_actors_retry_with_backoff_on_failure(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    failure_time, success_time = None, None
    succeeded = Event()

    # And an actor that fails the first time it's called
    @dramatiq.actor(min_backoff=1000, max_backoff=5000)
    def do_work():
        nonlocal failure_time, success_time
        if not failure_time:
            failure_time = current_millis()
            raise RuntimeError("First failure.")
        else:
            success_time = current_millis()
            succeeded.set()

    # If I send it a message
    do_work.send()

    # Then wait for the actor to succeed
    succeeded.wait(timeout=30)

    # I expect backoff time to have passed between success and failure
    assert 1000 <= success_time - failure_time <= (2000 + 200)
    # The first backoff time should be 100-200% of min_backoff.
    # Add an extra 200 milliseconds to account for processing and to prevent flakiness.


def test_rabbitmq_actors_can_retry_multiple_times(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    attempts = []

    # And an actor that fails 3 times then succeeds
    @dramatiq.actor(max_backoff=1000)
    def do_work():
        attempts.append(1)
        if sum(attempts) < 4:
            raise RuntimeError("Failure #%d" % sum(attempts))

    # If I send it a message
    do_work.send()

    # Then join on the queue
    rabbitmq_broker.join(do_work.queue_name, min_successes=40)
    rabbitmq_worker.join()

    # I expect it to have been attempted 4 times
    assert sum(attempts) == 4


def test_rabbitmq_actors_can_have_their_messages_delayed(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    start_time, run_time = current_millis(), None

    # And an actor that records the time it ran
    @dramatiq.actor
    def record():
        nonlocal run_time
        run_time = current_millis()

    # If I send it a delayed message
    record.send_with_options(delay=1000)

    # Then join on the queue
    rabbitmq_broker.join(record.queue_name)
    rabbitmq_worker.join()

    # I expect that message to have been processed at least delayed milliseconds later
    assert run_time - start_time >= 1000


def test_rabbitmq_actors_can_delay_messages_independent_of_each_other(rabbitmq_broker):
    # Given that I have a database
    results = []

    # And an actor that appends a number to the database
    @dramatiq.actor
    def append(x):
        results.append(x)

    # And a worker
    broker = rabbitmq_broker
    worker = Worker(broker, worker_threads=1)

    try:
        # And I send it a delayed message
        append.send_with_options(args=(1,), delay=1500)

        # And then another delayed message with a smaller delay
        append.send_with_options(args=(2,), delay=1000)

        # Then resume the worker and join on the queue
        worker.start()
        broker.join(append.queue_name, min_successes=20)
        worker.join()

        # I expect the latter message to have been run first
        assert results == [2, 1]
    finally:
        worker.stop()


def test_rabbitmq_actors_can_have_retry_limits(rabbitmq_broker, rabbitmq_worker):
    # Given that I have an actor that always fails
    @dramatiq.actor(max_retries=0)
    def do_work():
        raise RuntimeError("failed")

    # If I send it a message
    do_work.send()

    # Then join on its queue
    rabbitmq_broker.join(do_work.queue_name)
    rabbitmq_worker.join()

    # I expect the message to get moved to the dead letter queue
    _, _, xq_count = rabbitmq_broker.get_queue_message_counts(do_work.queue_name)
    assert xq_count == 1


@skip_unless_rabbit_mq
def test_rabbitmq_broker_connections_are_lazy():
    # When I create an RMQ broker
    broker = RabbitmqBroker(
        host="127.0.0.1",
        max_priority=10,
        credentials=RABBITMQ_CREDENTIALS,
    )

    def get_connection():
        return getattr(broker.state, "connection", None)

    # Then it shouldn't immediately connect to the server
    assert get_connection() is None

    # When I declare a queue
    broker.declare_queue("some-queue")

    # Then it shouldn't connect either
    assert get_connection() is None

    # When I create a consumer on that queue
    broker.consume("some-queue", timeout=1)

    # Then it should connect
    assert get_connection() is not None


def test_rabbitmq_broker_stops_retrying_declaring_queues_when_max_attempts_reached(rabbitmq_broker):
    # Given that I have a rabbit instance that lost its connection
    with patch.object(
        rabbitmq_broker,
        "_declare_queue",
        side_effect=pika.exceptions.AMQPConnectionError,
    ) as mock_declare_queue:
        # When I declare and use an actor
        # Then a ConnectionClosed error should be raised
        with pytest.raises(dramatiq.errors.ConnectionClosed):

            @dramatiq.actor(queue_name="flaky_queue")
            def do_work():
                pass

            do_work.send()

    # check declare was attempted the max number of times.
    assert mock_declare_queue.call_count == MAX_DECLARE_ATTEMPTS


def test_rabbitmq_messages_belonging_to_missing_actors_are_rejected(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a broker without actors
    # If I send it a message
    message = Message(
        queue_name="some-queue",
        actor_name="some-actor",
        args=(),
        kwargs={},
        options={},
    )
    rabbitmq_broker.declare_queue(message.queue_name)
    rabbitmq_broker.enqueue(message)

    # Then join on the queue
    rabbitmq_broker.join(message.queue_name)
    rabbitmq_worker.join()

    # I expect the message to end up on the dead letter queue
    _, _, dead = rabbitmq_broker.get_queue_message_counts(message.queue_name)
    assert dead == 1


def test_rabbitmq_broker_reconnects_after_enqueue_failure(rabbitmq_broker):
    # Given that I have an actor
    @dramatiq.actor
    def do_nothing():
        pass

    # If I close my connection
    rabbitmq_broker.connection.close()

    # Then send my actor a message
    # I expect the message to be enqueued
    assert do_nothing.send()

    # And the connection be reopened
    assert rabbitmq_broker.connection.is_open


def test_rabbitmq_workers_handle_rabbit_failures_gracefully(rabbitmq_broker, rabbitmq_worker):
    # Given that I have an attempts database
    attempts = []

    # And an actor that adds 1 to the attempts database
    @dramatiq.actor
    def do_work():
        attempts.append(1)
        time.sleep(1)

    # If I send that actor a delayed message
    do_work.send_with_options(delay=1000)

    # If I stop the RabbitMQ app
    os.system("rabbitmqctl stop_app")

    # Then start the app back up
    os.system("rabbitmqctl start_app")

    # And join on the queue
    del rabbitmq_broker.channel
    del rabbitmq_broker.connection
    rabbitmq_broker.join(do_work.queue_name)
    rabbitmq_worker.join()

    # I expect the work to have been attempted at least once
    assert sum(attempts) >= 1


def test_rabbitmq_connections_can_be_deleted_multiple_times(rabbitmq_broker):
    del rabbitmq_broker.connection
    del rabbitmq_broker.connection


def test_rabbitmq_channels_can_be_deleted_multiple_times(rabbitmq_broker):
    del rabbitmq_broker.channel
    del rabbitmq_broker.channel


def test_rabbitmq_consumers_ignore_unknown_messages_in_ack_and_nack(rabbitmq_broker):
    # Given that I have a RabbitmqConsumer
    consumer = rabbitmq_broker.consume("default")

    # If I attempt to ack a Message that wasn't consumed off of it
    # I expect nothing to happen
    assert consumer.ack(Mock(_tag=1)) is None

    # Likewise for nack
    assert consumer.nack(Mock(_tag=1)) is None


def test_ignore_scary_logs_filter_ignores_logs():
    # Given a filter that ignores scary logs
    log_filter = _IgnoreScaryLogs("pika.adapters")

    # When I ask it to filter a log message that contains a scary message
    record = Mock()
    record.getMessage.return_value = "ConnectionError('Broken pipe')"

    # Then it should filter out that log message
    assert not log_filter.filter(record)

    # And when I ask it to filter a log message that doesn't
    record = Mock()
    record.getMessage.return_value = "Not scary"

    # Then it should ignore that log message
    assert log_filter.filter(record)


def test_rabbitmq_close_only_registers_ignore_filter_once():
    # Given a RabbitmqBroker
    broker = RabbitmqBroker()

    base_logger = logging.getLogger("pika.adapters.base_connection")
    blocking_logger = logging.getLogger("pika.adapters.blocking_connection")

    # And snapshots of the current filters
    original_base_filters = list(base_logger.filters)
    original_blocking_filters = list(blocking_logger.filters)

    try:
        # When I close the broker twice
        broker.close()
        broker.close()

        base_filters = [f for f in base_logger.filters if isinstance(f, _IgnoreScaryLogs)]
        blocking_filters = [f for f in blocking_logger.filters if isinstance(f, _IgnoreScaryLogs)]

        # Then only one ignore filter is registered per logger
        assert len(base_filters) == 1
        assert len(blocking_filters) == 1
    finally:
        # And the filters are removed after the test run
        # so they don't affect the global state.
        for log_filter in list(base_logger.filters):
            if log_filter not in original_base_filters:
                base_logger.removeFilter(log_filter)

        for log_filter in list(blocking_logger.filters):
            if log_filter not in original_blocking_filters:
                blocking_logger.removeFilter(log_filter)


def test_rabbitmq_broker_can_join_with_timeout(rabbitmq_broker, rabbitmq_worker):
    # Given that I have an actor that takes a long time to run
    @dramatiq.actor
    def do_work():
        time.sleep(1)

    # When I send that actor a message
    do_work.send()

    # And join on its queue with a timeout
    # Then I expect a QueueJoinTimeout to be raised
    with pytest.raises(QueueJoinTimeout):
        rabbitmq_broker.join(do_work.queue_name, timeout=500)


def test_rabbitmq_broker_can_flush_queues(rabbitmq_broker):
    # Given that I have an actor
    @dramatiq.actor
    def do_work():
        pass

    # When I send that actor a message
    do_work.send()

    # And then tell the broker to flush all queues
    rabbitmq_broker.flush_all()

    # And then join on the actors's queue
    # Then it should join immediately
    assert rabbitmq_broker.join(do_work.queue_name, min_successes=1, timeout=200) is None


def test_rabbitmq_broker_can_enqueue_messages_with_priority(rabbitmq_broker):
    max_priority = 10
    message_processing_order = []
    queue_name = "prioritized"

    # Given that I have an actor that store priorities
    @dramatiq.actor(queue_name=queue_name)
    def do_work(message_priority):
        message_processing_order.append(message_priority)

    worker = Worker(rabbitmq_broker, worker_threads=1)
    worker.queue_prefetch = 1
    worker.start()
    worker.pause()

    try:
        # When I send that actor messages with increasing priorities
        for priority in range(max_priority):
            do_work.send_with_options(args=(priority,), broker_priority=priority)

        # And then tell the broker to wait for all messages
        worker.resume()
        rabbitmq_broker.join(queue_name, timeout=5000)
        worker.join()

        # I expect the stored priorities to be saved in decreasing order
        assert message_processing_order == list(reversed(range(max_priority)))
    finally:
        worker.stop()


def test_rabbitmq_broker_retries_declaring_queues_when_connection_related_errors_occur(rabbitmq_broker):
    executed, declare_called = False, False
    original_declare = rabbitmq_broker._declare_queue

    def flaky_declare_queue(*args, **kwargs):
        nonlocal declare_called
        if not declare_called:
            declare_called = True
            raise pika.exceptions.AMQPConnectionError
        return original_declare(*args, **kwargs)

    # Given that I have a flaky connection to a rabbitmq server
    with patch.object(rabbitmq_broker, "_declare_queue", flaky_declare_queue):
        # When I declare an actor
        @dramatiq.actor(queue_name="flaky_queue")
        def do_work():
            nonlocal executed
            executed = True

        # And I send that actor a message
        do_work.send()

        # And wait for the worker to process the message
        worker = Worker(rabbitmq_broker, worker_threads=1)
        worker.start()

        try:
            rabbitmq_broker.join(do_work.queue_name, timeout=5000)
            worker.join()

            # Then the queue should eventually be declared and the message executed
            assert declare_called
            assert executed
        finally:
            worker.stop()


def test_rabbitmq_broker_retries_declaring_queues_when_declared_queue_disappears(rabbitmq_broker):
    executed = False

    # Given that I have an actor on a flaky queue
    flaky_queue_name = "flaky_queue"
    rabbitmq_broker.channel.queue_delete(flaky_queue_name)

    @dramatiq.actor(queue_name=flaky_queue_name)
    def do_work():
        nonlocal executed
        executed = True

    # When I start a server
    worker = Worker(rabbitmq_broker, worker_threads=1)
    worker.start()

    declared_ev = Event()

    class DeclaredMiddleware(Middleware):
        def after_declare_queue(self, broker, queue_name):
            if queue_name == flaky_queue_name:
                declared_ev.set()

    # I expect that queue to be declared
    rabbitmq_broker.add_middleware(DeclaredMiddleware())
    assert declared_ev.wait(timeout=5)

    # If I delete the queue
    rabbitmq_broker.channel.queue_delete(do_work.queue_name)
    with pytest.raises(pika.exceptions.ChannelClosedByBroker):
        rabbitmq_broker.channel.queue_declare(do_work.queue_name, passive=True)

    # And I send that actor a message
    do_work.send()
    try:
        rabbitmq_broker.join(do_work.queue_name, timeout=20000)
        worker.join()
    finally:
        worker.stop()

    # Then the queue should be declared and the message executed
    assert executed


def test_rabbitmq_messages_that_failed_to_decode_are_rejected(rabbitmq_broker, rabbitmq_worker):
    # Given that I have an Actor
    @dramatiq.actor(max_retries=0)
    def do_work(_):
        pass

    old_encoder = dramatiq.get_encoder()

    # And an encoder that may fail to decode
    class BadEncoder(type(old_encoder)):
        def decode(self, data):
            if "xfail" in str(data):
                raise RuntimeError("xfail")
            return super().decode(data)

    dramatiq.set_encoder(BadEncoder())

    try:
        # When I send a message that will fail to decode
        do_work.send("xfail")

        # And I join on the queue
        rabbitmq_broker.join(do_work.queue_name)
        rabbitmq_worker.join()

        # Then I expect the message to get moved to the dead letter queue
        q_count, dq_count, xq_count = rabbitmq_broker.get_queue_message_counts(do_work.queue_name)

        assert q_count == dq_count == 0
        assert xq_count == 1
    finally:
        dramatiq.set_encoder(old_encoder)


def test_rabbitmq_queues_only_contains_canonical_name(rabbitmq_broker, rabbitmq_worker):
    assert len(rabbitmq_broker.queues) == 0

    @dramatiq.actor
    def put():
        pass

    assert len(rabbitmq_broker.queues) == 1
    assert put.queue_name in rabbitmq_broker.queues



from __future__ import annotations

import time

from dramatiq.rate_limits import BucketRateLimiter

from .common import skip_in_ci


@skip_in_ci
def test_bucket_rate_limiter_limits_per_bucket(rate_limiter_backend):
    # Given that I have a bucket rate limiter and a call database
    limiter = BucketRateLimiter(rate_limiter_backend, "sequential-test", limit=2)
    calls = 0

    for _ in range(2):
        # And I wait until the next second starts
        now = time.time()
        time.sleep(1 - (now - int(now)))

        # And I acquire it multiple times sequentially
        for _ in range(8):
            with limiter.acquire(raise_on_failure=False) as acquired:
                if not acquired:
                    continue

                calls += 1

    # I expect it to have succeeded four times
    assert calls == 4



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import re
import time
from datetime import timedelta
from inspect import iscoroutinefunction
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    ParamSpec,
    TypeVar,
    Union,
    overload,
)

from .asyncio import async_to_sync
from .broker import Broker, get_broker
from .logging import get_logger
from .message import Message

#: The regular expression that represents valid queue names.
_queue_name_re = re.compile(r"[a-zA-Z_][a-zA-Z0-9._-]*")

# Type variables for the Actor function's parameters and return type.
P = ParamSpec("P")
R = TypeVar("R")


class Actor(Generic[P, R]):
    """Thin wrapper around callables that stores metadata about how
    they should be executed asynchronously.  Actors are callable.

    Attributes:
      logger(Logger): The actor's logger.
      fn(callable): The underlying callable.
      broker(Broker): The broker this actor is bound to.
      actor_name(str): The actor's name.
      queue_name(str): The actor's queue.
      priority(int): The actor's priority.
      options(dict): Arbitrary options that are passed to the broker
        and middleware.
    """

    def __init__(
        self,
        fn: Callable[P, Union[R, Awaitable[R]]],
        *,
        broker: Broker,
        actor_name: str,
        queue_name: str,
        priority: int,
        options: dict[str, Any],
    ) -> None:
        if actor_name in broker.actors:
            raise ValueError(f"An actor named {actor_name!r} is already registered.")

        self.logger = get_logger(fn.__module__ or "_", actor_name)
        self.fn = async_to_sync(fn) if iscoroutinefunction(fn) else fn
        self.broker = broker
        self.actor_name = actor_name
        self.queue_name = queue_name
        self.priority = priority
        self.options = options
        self.broker.declare_actor(self)

    def message(self, *args: P.args, **kwargs: P.kwargs) -> Message[R]:
        """Build a message.  This method is useful if you want to
        compose actors.  See the actor composition documentation for
        details.

        Parameters:
          *args(tuple): Positional arguments to send to the actor.
          **kwargs: Keyword arguments to send to the actor.

        Examples:
          >>> (add.message(1, 2) | add.message(3))
          pipeline([add(1, 2), add(3)])

        Returns:
          Message: A message that can be enqueued on a broker.
        """
        return self.message_with_options(args=args, kwargs=kwargs)

    def message_with_options(
        self,
        *,
        args: tuple = (),
        kwargs: Optional[dict[str, Any]] = None,
        **options,
    ) -> Message[R]:
        """Build a message with an arbitrary set of processing options.
        This method is useful if you want to compose actors.  See the
        actor composition documentation for details.

        Parameters:
          args(tuple): Positional arguments that are passed to the actor.
          kwargs(dict): Keyword arguments that are passed to the actor.
          **options: Arbitrary options that are passed to the
            broker and any registered middleware.

        Returns:
          Message: A message that can be enqueued on a broker.
        """
        for name in ["on_failure", "on_success"]:
            callback = options.get(name)
            if isinstance(callback, Actor):
                options[name] = callback.actor_name

            elif not isinstance(callback, (type(None), str)):
                raise TypeError(name + " value must be an Actor")

        return Message(
            queue_name=self.queue_name,
            actor_name=self.actor_name,
            args=args,
            kwargs=kwargs or {},
            options=options,
        )

    def send(self, *args: P.args, **kwargs: P.kwargs) -> Message[R]:
        """Asynchronously send a message to this actor.

        Parameters:
          *args: Positional arguments to send to the actor.
          **kwargs: Keyword arguments to send to the actor.

        Returns:
          Message: The enqueued message.
        """
        return self.send_with_options(args=args, kwargs=kwargs)

    def send_with_options(
        self,
        *,
        args: tuple = (),
        kwargs: Optional[dict[str, Any]] = None,
        delay: Optional[timedelta | int] = None,
        **options,
    ) -> Message[R]:
        """Asynchronously send a message to this actor, along with an
        arbitrary set of processing options for the broker and
        middleware.

        Parameters:
          args(tuple): Positional arguments that are passed to the actor.
          kwargs(dict): Keyword arguments that are passed to the actor.
          delay(int): The minimum amount of time, in milliseconds, the
            message should be delayed by. Also accepts a timedelta.
          **options: Arbitrary options that are passed to the
            broker and any registered middleware.

        Returns:
          Message: The enqueued message.
        """
        if isinstance(delay, timedelta):
            delay = int(delay.total_seconds() * 1000)

        message = self.message_with_options(args=args, kwargs=kwargs, **options)
        return self.broker.enqueue(message, delay=delay)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Any | R | Awaitable[R]:
        """Synchronously call this actor.

        Parameters:
          *args: Positional arguments to send to the actor.
          **kwargs: Keyword arguments to send to the actor.

        Returns:
          Whatever the underlying function backing this actor returns.
        """
        try:
            self.logger.debug("Received args=%r kwargs=%r.", args, kwargs)
            start = time.perf_counter()
            return self.fn(*args, **kwargs)
        finally:
            delta = time.perf_counter() - start
            self.logger.debug("Completed after %.02fms.", delta * 1000)

    def __repr__(self) -> str:
        return "Actor(%(fn)r, queue_name=%(queue_name)r, actor_name=%(actor_name)r)" % vars(self)

    def __str__(self) -> str:
        return "Actor(%(actor_name)s)" % vars(self)


@overload
def actor(fn: Callable[P, Union[Awaitable[R], R]], **kwargs) -> Actor[P, R]:
    pass


@overload
def actor(fn: None = None, **kwargs) -> Callable[[Callable[P, Union[Awaitable[R], R]]], Actor[P, R]]:
    pass


def actor(
    fn: Optional[Callable[P, Union[Awaitable[R], R]]] = None,
    *,
    actor_class: Callable[..., Actor[P, R]] = Actor,
    actor_name: Optional[str] = None,
    queue_name: str = "default",
    priority: int = 0,
    broker: Optional[Broker] = None,
    **options,
) -> Union[Actor[P, R], Callable]:
    """Declare an actor.

    Examples:

      >>> import dramatiq

      >>> @dramatiq.actor
      ... def add(x, y):
      ...     print(x + y)
      ...
      >>> add
      Actor(<function add at 0x106c6d488>, queue_name='default', actor_name='add')

      >>> add(1, 2)
      3

      >>> add.send(1, 2)
      Message(
        queue_name='default',
        actor_name='add',
        args=(1, 2), kwargs={}, options={},
        message_id='e0d27b45-7900-41da-bb97-553b8a081206',
        message_timestamp=1497862448685)

    Parameters:
      fn(callable): The function to wrap.
      actor_class(type): Type created by the decorator.  Defaults to
        :class:`Actor` but can be any callable as long as it returns an
        actor and takes the same arguments as the :class:`Actor` class.
      actor_name(str): The name of the actor.
      queue_name(str): The name of the queue to use.
      priority(int): The actor's global priority.  If two tasks have
        been pulled on a worker concurrently and one has a higher
        priority than the other then it will be processed first.
        Lower numbers represent higher priorities.
      broker(Broker): The broker to use with this actor.
      **options: Arbitrary options that vary with the set of
        middleware that you use.  See ``get_broker().actor_options``.

    Returns:
      Actor: The decorated function.
    """

    def decorator(fn: Callable[..., Union[Awaitable[R], R]]) -> Actor[P, R]:
        nonlocal actor_name, broker
        actor_name = actor_name or fn.__name__
        if not _queue_name_re.fullmatch(queue_name):
            raise ValueError(
                "Queue names must start with a letter or an underscore followed "
                "by any number of letters, digits, dashes or underscores."
            )

        broker = broker or get_broker()
        invalid_options = set(options) - broker.actor_options
        if invalid_options:
            invalid_options_list = ", ".join(invalid_options)
            raise ValueError(
                "The following actor options are undefined: %s. Did you forget to add a middleware to your Broker?"
                % invalid_options_list
            )

        return actor_class(
            fn,
            actor_name=actor_name,
            queue_name=queue_name,
            priority=priority,
            broker=broker,
            options=options,
        )

    if fn is None:
        return decorator
    return decorator(fn)



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import dataclasses
import time
import uuid
from datetime import datetime, timezone
from typing import Any, Generic, Optional, TypeVar

from .broker import get_broker
from .composition import pipeline
from .encoder import Encoder, JSONEncoder
from .errors import DecodeError
from .results import ResultBackend

#: The global encoder instance.
global_encoder: Encoder = JSONEncoder()

R = TypeVar("R", covariant=True)


def get_encoder() -> Encoder:
    """Get the global encoder object.

    Returns:
      Encoder
    """
    return global_encoder


def set_encoder(encoder: Encoder) -> None:
    """Set the global encoder object.

    Parameters:
      encoder(Encoder): The encoder instance to use when serializing
        messages.
    """
    global global_encoder
    global_encoder = encoder


def generate_unique_id() -> str:
    """Generate a globally-unique message id."""
    return str(uuid.uuid4())


@dataclasses.dataclass(frozen=True)
class Message(Generic[R]):
    """Encapsulates metadata about messages being sent to individual actors.

    Parameters:
      queue_name(str): The name of the queue the message belongs to.
      actor_name(str): The name of the actor that will receive the message.
      args(tuple): Positional arguments that are passed to the actor.
      kwargs(dict): Keyword arguments that are passed to the actor.
      options(dict): Arbitrary options passed to the broker and middleware.
      message_id(str): A globally-unique id assigned to the actor.
      message_timestamp(int): The UNIX timestamp in milliseconds
        representing when the message was first enqueued.
    """

    queue_name: str
    actor_name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    options: dict[str, Any]
    message_id: str = dataclasses.field(default_factory=generate_unique_id)
    message_timestamp: int = dataclasses.field(default_factory=lambda: int(time.time() * 1000))

    def __post_init__(self):
        # For backwards-compatibility, enforce that `args' is a tuple.
        if type(self.args) is not tuple:
            # The class is marked frozen, so we have to use the
            # primitive setattr here.  Direct assignment would fail.
            object.__setattr__(self, "args", tuple(self.args))

    def __or__(self, other) -> pipeline:
        """Combine this message into a pipeline with "other"."""
        return pipeline([self, other])

    def asdict(self) -> dict[str, Any]:
        """Convert this message to a dictionary."""
        # For backward compatibility, we can't use `dataclasses.asdict`
        # because it creates a copy of all values, including `options`.
        result = {}
        for field in dataclasses.fields(self):
            result[field.name] = getattr(self, field.name)
        return result

    @classmethod
    def decode(cls, data: bytes) -> Message:
        """Convert a bytestring to a message.

        Raises:
          DecodeError: When the decoder raises an exception while
            decoding `data`.
        """
        try:
            fields = global_encoder.decode(data)
            fields["args"] = tuple(fields["args"])
            return cls(**fields)
        except Exception as e:
            raise DecodeError("Failed to decode message.", data, e) from e

    def encode(self) -> bytes:
        """Convert this message to a bytestring."""
        return global_encoder.encode(self.asdict())

    def copy(self, **attributes) -> Message:
        """Create a copy of this message."""
        new_options = attributes.pop("options", {})
        return dataclasses.replace(self, **attributes, options={**self.options, **new_options})

    def get_result(
        self,
        *,
        backend: Optional[ResultBackend] = None,
        block: bool = False,
        timeout: Optional[int] = None,
    ) -> R:
        """Get the result associated with this message from a result
        backend.

        Warning:
          If you use multiple result backends or brokers you should
          always pass the backend parameter.  This method is only able
          to infer the result backend off of the default broker.

        Parameters:
          backend(ResultBackend): The result backend to use to get the
            result.  If omitted, this method will try to find and use
            the result backend on the default broker instance.
          block(bool): Whether or not to block while waiting for a
            result.
          timeout(int): The maximum amount of time, in ms, to block
            while waiting for a result.  Defaults to 10 seconds.

        Raises:
          RuntimeError: If there is no result backend on the default
            broker.
          ResultMissing: When block is False and the result isn't set.
          ResultTimeout: When waiting for a result times out.

        Returns:
          object: The result.
        """
        if backend is None:
            broker = get_broker()
            backend = broker.get_results_backend()

        return backend.get_result(self, block=block, timeout=timeout)

    def __str__(self) -> str:
        params = ", ".join(repr(arg) for arg in self.args)
        if self.kwargs:
            params += ", " if params else ""
            params += ", ".join("%s=%r" % (name, value) for name, value in self.kwargs.items())

        return "%s(%s)" % (self.actor_name, params)

    def __lt__(self, other: Message) -> bool:
        return dataclasses.astuple(self) < dataclasses.astuple(other)

    # Backwards-compatibility with namedtuple.
    _asdict = asdict

    @property
    def _field_defaults(self) -> dict[str, Any]:
        return {f.name: f.default for f in dataclasses.fields(self) if f.default is not dataclasses.MISSING}

    @property
    def _fields(self) -> tuple[str, ...]:
        return tuple(f.name for f in dataclasses.fields(self))

    def _replace(self, **changes) -> Message[R]:
        return dataclasses.replace(self, **changes)

    @property
    def message_datetime(self) -> datetime:
        """Read ``message_timestamp`` as a UTC-aware datetime.

        Datetime precision is limited by the representation of ``Message.message_timestamp``, which is an
        integer storing milliseconds.
        """
        unix_seconds, ms = divmod(self.message_timestamp, 1000)
        return datetime.fromtimestamp(unix_seconds, tz=timezone.utc).replace(microsecond=ms * 1000)



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import dramatiq


def test_messages_have_namedtuple_methods(stub_broker):
    @dramatiq.actor
    def add(x, y):
        return x + y

    msg1 = add.message(1, 2)
    assert msg1.asdict() == msg1._asdict()

    assert msg1._field_defaults == {}
    assert msg1._fields == (
        "queue_name",
        "actor_name",
        "args",
        "kwargs",
        "options",
        "message_id",
        "message_timestamp",
    )

    msg2 = msg1._replace(queue_name="example")
    assert msg2._asdict()["queue_name"] == "example"


def test_messageproxy_representation(stub_broker):
    """Ensure that MessageProxy defines ``__repr__``.

    While Dramatiq logs messages and MessageProxies formatted using the "%s" placeholder, other tools like Sentry-SDK
    force ``repr`` of any log param that isn't a primitive. This means that MessageProxy has to implement ``__repr__``
    in order for the captured messages to be more helpful than:

    > Failed to process message <dramatiq.broker.MessageProxy object at 0x74522262a950> with unhandled exception.
    """

    @dramatiq.actor
    def actor(arg):
        return arg

    message = actor.send("input")

    consumer = stub_broker.consume("default")
    message_proxy = next(consumer)

    assert repr(message) in repr(message_proxy), "Expecting MessageProxy repr to contain Message repr"
    assert str(message) == str(message_proxy), "Expecting identical __str__ of MessageProxy and Message"


def test_message_datetime(stub_broker):
    """Test reading message time as datetime."""

    @dramatiq.actor
    def actor(arg):
        return arg

    message = actor.send("input")
    after = datetime.now(timezone.utc)

    assert message.message_datetime.tzinfo is timezone.utc
    assert message.message_datetime < after
    assert message.message_datetime > after - timedelta(seconds=1)

    # no rounding problems here because message_timestamp is an int
    assert message.message_datetime.timestamp() == message.message_timestamp / 1000



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import threading
import warnings
from threading import Thread
from time import monotonic, sleep
from typing import TYPE_CHECKING, Optional, Union

from ..logging import get_logger
from .middleware import Middleware
from .threading import (
    Interrupt,
    current_platform,
    is_gevent_active,
    raise_thread_exception,
    supported_platforms,
)

if TYPE_CHECKING:
    import gevent


class TimeLimitExceeded(Interrupt):
    """Exception used to interrupt worker threads when actors exceed
    their time limits.
    """


class TimeLimit(Middleware):
    """Middleware that cancels actors that run for too long.
    Currently, this is only available on CPython.

    Note:
      This works by setting an async exception in the worker thread
      that runs the actor.  This means that the exception will only get
      called the next time that thread acquires the GIL.  Concretely,
      this means that this middleware can't cancel system calls.

    Parameters:
      time_limit(float): The maximum number of milliseconds actors may
        run for. Use `float("inf")` to avoid setting a timeout for the
        actor.
        Defaults to 10 minutes (600,000 milliseconds).
      interval(int): The interval (in milliseconds) with which to
        check for actors that have exceeded the limit. This does not take
        effect when using gevent because the timers are managed by gevent.
        Defaults to 1 second (1,000 milliseconds).
    """

    def __init__(self, *, time_limit: float = 600000, interval: int = 1000) -> None:
        self.logger = get_logger(__name__, type(self))
        self.time_limit = time_limit

        self.manager: Union[_GeventTimeoutManager, _CtypesTimeoutManager]
        if is_gevent_active():
            self.manager = _GeventTimeoutManager(logger=self.logger)
        else:
            self.manager = _CtypesTimeoutManager(interval, logger=self.logger)

    @property
    def actor_options(self):
        return {"time_limit"}

    def after_process_boot(self, broker):
        if is_gevent_active() or current_platform in supported_platforms:
            self.manager.start()

        else:  # pragma: no cover
            msg = "TimeLimit cannot kill threads on your current platform (%r)."
            warnings.warn(msg % current_platform, category=RuntimeWarning, stacklevel=2)

    def before_process_message(self, broker, message):
        actor = broker.get_actor(message.actor_name)
        limit = message.options.get("time_limit") or actor.options.get("time_limit", self.time_limit)
        self.manager.add_timeout(threading.get_ident(), limit)

    def after_process_message(self, broker, message, *, result=None, exception=None):
        self.manager.remove_timeout(threading.get_ident())

    after_skip_message = after_process_message


class _CtypesTimeoutManager(Thread):
    def __init__(self, interval, logger=None):
        super().__init__(daemon=True)
        self.deadlines = {}
        self.interval = interval / 1000
        self.logger = logger or get_logger(__name__, type(self))
        self.mu = threading.RLock()

    def _handle_deadlines(self):
        current_time = monotonic()
        threads_to_kill = []
        with self.mu:
            for thread_id, deadline in self.deadlines.items():
                if deadline and current_time >= deadline:
                    self.logger.warning(
                        "Time limit exceeded. Raising exception in worker thread %r.",
                        thread_id,
                    )
                    self.deadlines[thread_id] = None
                    threads_to_kill.append(thread_id)

        for thread_id in threads_to_kill:
            raise_thread_exception(thread_id, TimeLimitExceeded)

    def run(self):
        while True:
            try:
                self._handle_deadlines()
            except Exception:  # pragma: no cover
                self.logger.exception("Unhandled error while running the time limit handler.")

            sleep(self.interval)

    def add_timeout(self, thread_id, ttl):
        with self.mu:
            self.deadlines[thread_id] = monotonic() + ttl / 1000

    def remove_timeout(self, thread_id):
        with self.mu:
            self.deadlines[thread_id] = None


class _GeventTimeoutManager:
    def __init__(self, logger=None):
        self.timers = {}
        self.logger = logger or get_logger(__name__, type(self))

    def start(self):
        pass

    def add_timeout(self, thread_id, ttl):
        self.timers[thread_id] = _GeventTimeout(
            logger=self.logger,
            thread_id=thread_id,
            after_expiration=lambda: self.timers.pop(thread_id, None),
            seconds=None if ttl == float("inf") else ttl / 1000,
            exception=TimeLimitExceeded,
        )
        self.timers[thread_id].start()

    def remove_timeout(self, thread_id):
        timer = self.timers.pop(thread_id, None)
        if timer is not None:
            timer.close()


_GeventTimeout: Optional[gevent.Timeout] = None
if is_gevent_active():
    from gevent import Timeout

    class __GeventTimeout(Timeout):
        """Cooperative timeout class for gevent with logging on timeouts."""

        def __init__(self, *args, logger=None, thread_id=None, after_expiration=None, **kwargs):
            super().__init__(*args, **kwargs)
            self.logger = logger or get_logger(__name__, type(self))
            self.thread_id = thread_id
            self.after_expiration = after_expiration

        def _on_expiration(self, prev_greenlet, ex):
            self.logger.warning(
                "Time limit exceeded. Raising exception in worker thread %r.",
                self.thread_id,
            )
            res = super()._on_expiration(prev_greenlet, ex)
            if self.after_expiration is not None:
                self.after_expiration()
            return res

    _GeventTimeout = __GeventTimeout



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os
import time
from functools import partial
from itertools import chain
from threading import Event, local
from typing import Any, Optional, Union

import pika

from ..broker import Broker, Consumer, MessageProxy
from ..common import current_millis, dq_name, q_name, xq_name
from ..errors import ConnectionClosed, DecodeError, QueueJoinTimeout
from ..logging import get_logger
from ..message import Message, get_encoder
from ..middleware import Middleware

#: The maximum amount of time a message can be in the dead letter queue.
DEAD_MESSAGE_TTL = int(os.getenv("dramatiq_dead_message_ttl", 86400000 * 7))

#: The max number of times to attempt an enqueue operation in case of
#: a connection error.
MAX_ENQUEUE_ATTEMPTS = 6
MAX_DECLARE_ATTEMPTS = 2


class RabbitmqBroker(Broker):
    """A broker that can be used with RabbitMQ.

    Examples:

      If you want to specify connection parameters individually:

      >>> RabbitmqBroker(host="127.0.0.1", port=5672)

      Alternatively, if you want to use a connection URL:

      >>> RabbitmqBroker(url="amqp://guest:guest@127.0.0.1:5672")

      To support queued message priorities, provide a ``max_priority``...

      >>> broker = RabbitmqBroker(url="...", max_priority=5)

      ... then enqueue messages with the ``broker_priority`` option:

      >>> broker.enqueue(an_actor.message_with_options(
      ...    broker_priority=5,
      ... ))

      ``broker_priority`` can also be provided to ``send_with_options``:

      >>> an_actor.send_with_options(
      ...    broker_priority=5,
      ... )

    The ``broker_priority`` provided should have a value between 0 and ``max_priority``, inclusive.
    Messages without a priority are treated as priority 0.
    RabbitMQ treats higher numbers as higher priorities.

    Note:
        This the **opposite** to the Dramatiq actor ``priority`` option.
        (where lower numbers are higher priorities).

    See also:
      ConnectionParameters_ for a list of all the available connection
      parameters.

    Parameters:
      confirm_delivery(bool): Wait for RabbitMQ to confirm that
        messages have been committed on every call to enqueue.
        Defaults to False.
      url(str|list[str]): An optional connection URL.  If both a URL
        and connection parameters are provided, the URL is used.
      middleware(list[Middleware]): The set of middleware that apply
        to this broker.
      max_priority(int): Configure queues with ``x-max-priority`` to
        support queue-global priority queueing.
      parameters(list[dict]): A sequence of (pika) connection parameters
        to determine which Rabbit server(s) to connect to.
      **kwargs: The (pika) connection parameters to use to
        determine which Rabbit server to connect to.

    .. _ConnectionParameters: https://pika.readthedocs.io/en/stable/modules/parameters.html
    """

    def __init__(
        self,
        *,
        confirm_delivery: bool = False,
        url: Optional[Union[str, list[str]]] = None,
        middleware: Optional[list[Middleware]] = None,
        max_priority: Optional[int] = None,
        parameters: Optional[list[dict[str, Any]]] = None,
        **kwargs: Any,
    ):
        super().__init__(middleware=middleware)

        if max_priority is not None and not (0 < max_priority <= 255):
            raise ValueError("max_priority must be a value between 0 and 255")

        if url is not None:
            if parameters is not None or kwargs:
                raise RuntimeError("the 'url' argument cannot be used in conjunction with pika parameters")

            if isinstance(url, str) and ";" in url:
                self.parameters = [pika.URLParameters(u) for u in url.split(";")]

            elif isinstance(url, list):
                self.parameters = [pika.URLParameters(u) for u in url]

            else:
                self.parameters = pika.URLParameters(url)

        elif parameters is not None:
            if kwargs:
                raise RuntimeError("the 'parameters' argument cannot be used in conjunction with other pika parameters")

            self.parameters = [pika.ConnectionParameters(**p) for p in parameters]

        else:
            self.parameters = pika.ConnectionParameters(**kwargs)

        self.confirm_delivery = confirm_delivery
        self.max_priority = max_priority
        self.connections: set[pika.BlockingConnection] = set()
        self.channels: set[pika.BlockingChannel] = set()
        self.queues: set[str] = set()
        self.queues_pending: set[str] = set()
        self.state = local()

    @property
    def consumer_class(self):
        return _RabbitmqConsumer

    @property
    def connection(self):
        """The :class:`pika.BlockingConnection` for the current
        thread.  This property may change without notice.
        """
        connection = getattr(self.state, "connection", None)
        if connection is None:
            connection = self.state.connection = pika.BlockingConnection(parameters=self.parameters)
            self.connections.add(connection)
        return connection

    @connection.deleter
    def connection(self):
        del self.channel
        try:
            connection = self.state.connection
        except AttributeError:
            return

        del self.state.connection
        self.connections.remove(connection)
        if connection.is_open:
            try:
                connection.close()
            except Exception:
                self.logger.exception("Encountered exception while closing Connection.")

    @property
    def channel(self):
        """The :class:`pika.BlockingChannel` for the current thread.
        This property may change without notice.
        """
        channel = getattr(self.state, "channel", None)
        if channel is None:
            channel = self.state.channel = self.connection.channel()
            if self.confirm_delivery:
                channel.confirm_delivery()

            self.channels.add(channel)
        return channel

    @channel.deleter
    def channel(self):
        try:
            channel = self.state.channel
        except AttributeError:
            return

        del self.state.channel
        self.channels.remove(channel)
        if channel.is_open:
            try:
                channel.close()
            except Exception:
                self.logger.exception("Encountered exception while closing Channel.")

    def _ignore_pika_logs(self) -> None:
        """Ensures that pika logs are filtered.

        The main thread may keep connections open for a long time
        w/o publishing heartbeats, which means that they'll end up
        being closed by the time the broker is closed.  When that
        happens, pika logs a bunch of scary stuff so we want to
        filter that out.
        """

        logging_filter = _IgnoreScaryLogs()
        ignored_loggers = ["pika.adapters.base_connection", "pika.adapters.blocking_connection"]

        # Make sure the filter is added only once.
        for logger_name in ignored_loggers:
            ignored_logger = logging.getLogger(logger_name)
            if not any(isinstance(f, _IgnoreScaryLogs) for f in ignored_logger.filters):
                ignored_logger.addFilter(logging_filter)

    def close(self) -> None:
        """Close all open RabbitMQ connections."""

        self._ignore_pika_logs()

        self.logger.debug("Closing channels and connections...")
        for channel_or_conn in chain(self.channels, self.connections):
            try:
                channel_or_conn.close()
            except pika.exceptions.AMQPError:
                pass

            except Exception:  # pragma: no cover
                self.logger.debug(
                    "Encountered an error while closing %r.",
                    channel_or_conn,
                    exc_info=True,
                )
        self.logger.debug("Channels and connections closed.")

    def consume(self, queue_name: str, prefetch: int = 1, timeout: int = 5000) -> Consumer:
        """Create a new consumer for a queue.

        Parameters:
          queue_name(str): The queue to consume.
          prefetch(int): The number of messages to prefetch.
          timeout(int): The idle timeout in milliseconds.

        Returns:
          Consumer: A consumer that retrieves messages from RabbitMQ.
        """
        self.declare_queue(queue_name, ensure=True)
        return self.consumer_class(self.parameters, queue_name, prefetch, timeout)

    def declare_queue(self, queue_name: str, *, ensure: bool = False) -> None:
        """Declare a queue.  Has no effect if a queue with the given
        name already exists.

        Parameters:
          queue_name(str): The name of the new queue.
          ensure(bool): When True, the queue is created on the server,
            if necessary.

        Raises:
          ConnectionClosed: When ensure=True if the underlying channel
            or connection fails.
        """
        if q_name(queue_name) not in self.queues:
            self.emit_before("declare_queue", queue_name)
            self.queues.add(queue_name)
            self.queues_pending.add(queue_name)
            self.emit_after("declare_queue", queue_name)

            delayed_name = dq_name(queue_name)
            self.delay_queues.add(delayed_name)
            self.emit_after("declare_delay_queue", delayed_name)

        if ensure:
            self._ensure_queue(queue_name)

    def _ensure_queue(self, queue_name):
        attempts = 0
        while True:
            try:
                if queue_name in self.queues_pending:
                    self._declare_queue(queue_name)
                    self._declare_dq_queue(queue_name)
                    self._declare_xq_queue(queue_name)
                    self.queues_pending.discard(queue_name)

                break
            except (
                pika.exceptions.AMQPConnectionError,
                pika.exceptions.AMQPChannelError,
            ) as e:  # pragma: no cover
                # Delete the channel and the connection so that the next
                # caller may initiate new ones of each.
                del self.connection

                attempts += 1
                if attempts >= MAX_DECLARE_ATTEMPTS:
                    raise ConnectionClosed(e) from None

                self.logger.debug(
                    "Retrying declare due to closed connection. [%d/%d] attempts made so far.",
                    attempts,
                    MAX_DECLARE_ATTEMPTS,
                )

    def _build_queue_arguments(self, queue_name):
        arguments = {
            "x-dead-letter-exchange": "",
            "x-dead-letter-routing-key": xq_name(queue_name),
        }
        if self.max_priority:
            arguments["x-max-priority"] = self.max_priority

        return arguments

    def _declare_queue(self, queue_name):
        arguments = self._build_queue_arguments(queue_name)
        return self.channel.queue_declare(queue=queue_name, durable=True, arguments=arguments)

    def _declare_dq_queue(self, queue_name):
        arguments = self._build_queue_arguments(queue_name)
        return self.channel.queue_declare(queue=dq_name(queue_name), durable=True, arguments=arguments)

    def _declare_xq_queue(self, queue_name):
        return self.channel.queue_declare(
            queue=xq_name(queue_name),
            durable=True,
            arguments={
                # This HAS to be a static value since messages are expired
                # in order inside of RabbitMQ (head-first).
                "x-message-ttl": DEAD_MESSAGE_TTL,
            },
        )

    def enqueue(self, message: Message, *, delay: Optional[int] = None) -> Message:
        """Enqueue a message.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The minimum amount of time, in milliseconds, to
            delay the message by.

        Raises:
          ConnectionClosed: If the underlying channel or connection
            has been closed.
        """
        canonical_queue_name = message.queue_name
        queue_name = canonical_queue_name

        if delay is not None:
            queue_name = dq_name(queue_name)
            message_eta = current_millis() + delay
            message = message.copy(
                queue_name=queue_name,
                options={
                    "eta": message_eta,
                },
            )

        attempts = 0
        while True:
            try:
                self.declare_queue(canonical_queue_name, ensure=True)
                self.logger.debug("Enqueueing message %r on queue %r.", message.message_id, queue_name)
                self.emit_before("enqueue", message, delay)
                self.channel.basic_publish(
                    exchange="",
                    routing_key=queue_name,
                    body=message.encode(),
                    properties=pika.BasicProperties(
                        delivery_mode=2,
                        priority=message.options.get("broker_priority"),
                    ),
                )
                self.emit_after("enqueue", message, delay)
                return message

            except (
                pika.exceptions.AMQPConnectionError,
                pika.exceptions.AMQPChannelError,
            ) as e:
                # Delete the channel and the connection so that the
                # next caller/attempt may initiate new ones of each.
                del self.connection

                # If the queue disappears, remove it from the known set
                # so that it can be redeclared on retry or the next time
                # a message is enqueued.
                if getattr(e, "reply_code", None) == 404:
                    self.queues.remove(q_name(queue_name))

                attempts += 1
                if attempts >= MAX_ENQUEUE_ATTEMPTS:
                    raise ConnectionClosed(e) from None

                self.logger.debug(
                    "Retrying enqueue due to closed connection. [%d/%d] attempts made so far.",
                    attempts,
                    MAX_ENQUEUE_ATTEMPTS,
                )

    def get_declared_queues(self) -> set[str]:
        """Get all declared queues.

        Returns:
          set[str]: The names of all the queues declared so far on
          this Broker.
        """
        return self.queues.copy()

    def get_queue_message_counts(self, queue_name: str) -> tuple[int, int, int]:
        """Get the number of messages in a queue.  This method is only
        meant to be used in unit and integration tests.

        Parameters:
          queue_name(str): The queue whose message counts to get.

        Returns:
          tuple: A triple representing the number of messages in the
          queue, its delayed queue and its dead letter queue.
        """
        queue_response = self._declare_queue(queue_name)
        dq_queue_response = self._declare_dq_queue(queue_name)
        xq_queue_response = self._declare_xq_queue(queue_name)
        return (
            queue_response.method.message_count,
            dq_queue_response.method.message_count,
            xq_queue_response.method.message_count,
        )

    def flush(self, queue_name: str) -> None:
        """Drop all the messages from a queue.

        Parameters:
          queue_name(str): The queue to flush.
        """
        # Purge messages from all queues, even from delayed queues and dead letter queues.
        # The purge operation fails with an exception if the queue doesn't exist. The the underlying
        # RabbitMQ channel is closed by the broker. We have to reopen the channel to continue
        # purging other queues.
        for name in (queue_name, dq_name(queue_name), xq_name(queue_name)):
            try:
                self.channel.queue_purge(name)
            except pika.exceptions.AMQPChannelError:
                del self.channel

    def flush_all(self) -> None:
        """Drop all messages from all declared queues."""
        for queue_name in self.queues:
            self.flush(queue_name)

    def join(
        self, queue_name: str, min_successes: int = 10, idle_time: int = 100, *, timeout: Optional[int] = None
    ) -> None:
        """Wait for all the messages on the given queue to be
        processed.  This method is only meant to be used in tests to
        wait for all the messages in a queue to be processed.

        Warning:
          This method doesn't wait for unacked messages so it may not
          be completely reliable.  Use the stub broker in your unit
          tests and only use this for simple integration tests.

        Parameters:
          queue_name(str): The queue to wait on.
          min_successes(int): The minimum number of times all the
            polled queues should be empty.
          idle_time(int): The number of milliseconds to wait between
            counts.
          timeout(Optional[int]): The max amount of time, in
            milliseconds, to wait on this queue.
        """
        deadline = timeout and time.monotonic() + timeout / 1000
        successes = 0
        while successes < min_successes:
            if deadline and time.monotonic() >= deadline:
                raise QueueJoinTimeout(queue_name)

            total_messages = sum(self.get_queue_message_counts(queue_name)[:-1])
            if total_messages == 0:
                successes += 1
            else:
                successes = 0

            self.connection.sleep(idle_time / 1000)


class _IgnoreScaryLogs(logging.Filter):
    def filter(self, record):
        return "Broken pipe" not in record.getMessage()


class _RabbitmqConsumer(Consumer):
    def __init__(self, parameters, queue_name, prefetch, timeout):
        try:
            self.logger = get_logger(__name__, type(self))
            self.connection = pika.BlockingConnection(parameters=parameters)
            self.channel = self.connection.channel()
            self.channel.basic_qos(prefetch_count=prefetch)
            self.iterator = self.channel.consume(queue_name, inactivity_timeout=timeout / 1000)

            # We need to keep track of known delivery tags so that
            # when connection errors occur and the consumer is reset,
            # we don't attempt to send invalid tags to Rabbit since
            # pika doesn't handle this very well.
            self.known_tags = set()
        except (
            pika.exceptions.AMQPConnectionError,
            pika.exceptions.AMQPChannelError,
        ) as e:
            raise ConnectionClosed(e) from None

    def ack(self, message):
        try:
            self.known_tags.remove(message._tag)
            self.connection.add_callback_threadsafe(
                partial(self.channel.basic_ack, message._tag),
            )
        except (
            pika.exceptions.AMQPConnectionError,
            pika.exceptions.AMQPChannelError,
        ) as e:
            raise ConnectionClosed(e) from None
        except KeyError:
            self.logger.warning("Failed to ack message: not in known tags.")
        except Exception:  # pragma: no cover
            self.logger.warning("Failed to ack message.", exc_info=True)

    def nack(self, message):
        try:
            self.known_tags.remove(message._tag)
            self._nack(message._tag)
        except (
            pika.exceptions.AMQPConnectionError,
            pika.exceptions.AMQPChannelError,
        ) as e:
            raise ConnectionClosed(e) from None
        except KeyError:
            self.logger.warning("Failed to nack message: not in known tags.")
        except Exception:  # pragma: no cover
            self.logger.warning("Failed to nack message.", exc_info=True)

    def _nack(self, tag):
        self.connection.add_callback_threadsafe(
            partial(self.channel.basic_nack, tag, requeue=False),
        )

    def requeue(self, messages):
        """RabbitMQ automatically re-enqueues unacked messages when
        consumers disconnect so this is a no-op.
        """

    def __next__(self):
        try:
            method, properties, body = next(self.iterator)
            if method is None:
                return None
        except (
            AssertionError,
            pika.exceptions.AMQPConnectionError,
            pika.exceptions.AMQPChannelError,
        ) as e:
            raise ConnectionClosed(e) from None

        try:
            message = Message.decode(body)
        except DecodeError:
            self.logger.exception("Failed to decode message using encoder %r.", get_encoder())
            self._nack(method.delivery_tag)
            return None

        rmq_message = _RabbitmqMessage(
            method.redelivered,
            method.delivery_tag,
            message,
        )
        self.known_tags.add(method.delivery_tag)
        return rmq_message

    def close(self):
        try:
            # Closing the connection doesn't wait for all callbacks to
            # finish processing so we enqueue a final callback and
            # wait for it to finish before closing the connection.
            # Assumes callbacks are called in order (they should be).
            all_callbacks_handled = Event()
            self.connection.add_callback_threadsafe(all_callbacks_handled.set)
            while not all_callbacks_handled.is_set():
                self.connection.sleep(0)
        except Exception:
            self.logger.exception(
                "Failed to wait for all callbacks to complete.  This "
                "can happen when the RabbitMQ server is suddenly "
                "restarted."
            )

        try:
            self.channel.close()
            self.connection.close()
        except (
            AssertionError,
            pika.exceptions.AMQPConnectionError,
            pika.exceptions.AMQPChannelError,
        ) as e:
            raise ConnectionClosed(e) from None


class _RabbitmqMessage(MessageProxy):
    def __init__(self, redelivered, tag, message):
        super().__init__(message)

        self.redelivered = redelivered
        self._tag = tag



# This file is a part of Dramatiq.
#
# Copyright (C) 2017,2018,2019 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Dramatiq is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Dramatiq is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os

from setuptools import setup


def rel(*xs):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *xs)


with open(rel("README.md")) as f:
    long_description = f.read()


with open(rel("dramatiq", "__init__.py"), "r") as f:
    version_marker = "__version__ = "
    for line in f:
        if line.startswith(version_marker):
            _, version = line.split(version_marker)
            version = version.strip().strip('"')
            break
    else:
        raise RuntimeError("Version marker not found.")


extra_dependencies = {
    "gevent": [
        "gevent>=1.1",
    ],
    "memcached": [
        "pylibmc>=1.5,<2.0",
    ],
    "prometheus": [
        "prometheus-client>=0.2",
    ],
    "rabbitmq": [
        "pika>=1.0,<2.0",
    ],
    "redis": [
        "redis>=4.0,<7.0",
    ],
    "watch": [
        "watchdog>=4.0",
        "watchdog_gevent>=0.2",
    ],
}

extra_dependencies["all"] = list(set(sum(extra_dependencies.values(), [])))

setup(
    name="dramatiq",
    version=version,
    author="Bogdan Popa",
    author_email="bogdan@cleartype.io",
    project_urls={
        "Documentation": "https://dramatiq.io",
        "Changelog": "https://dramatiq.io/changelog.html",
        "Source": "https://github.com/Bogdanp/dramatiq",
    },
    description="Background Processing for Python 3.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        "dramatiq",
        "dramatiq.brokers",
        "dramatiq.middleware",
        "dramatiq.rate_limits",
        "dramatiq.rate_limits.backends",
        "dramatiq.results",
        "dramatiq.results.backends",
    ],
    include_package_data=True,
    python_requires=">=3.10",
    extras_require=extra_dependencies,
    entry_points={"console_scripts": ["dramatiq = dramatiq.__main__:main"]},
    scripts=["bin/dramatiq-gevent"],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: System :: Distributed Computing",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    ],
)



from __future__ import annotations

import time
from unittest.mock import Mock

import pytest

import dramatiq
from dramatiq import QueueJoinTimeout, QueueNotFound


def test_stub_broker_raises_queue_error_when_consuming_undeclared_queues(stub_broker):
    # Given that I have a stub broker
    # If I attempt to consume a queue that wasn't declared
    # I expect a QueueNotFound error to be raised
    with pytest.raises(QueueNotFound):
        stub_broker.consume("idontexist")


def test_stub_broker_raises_queue_error_when_enqueueing_messages_on_undeclared_queues(stub_broker):
    # Given that I have a stub broker
    # If I attempt to enqueue a message on a queue that wasn't declared
    # I expect a QueueNotFound error to be raised
    with pytest.raises(QueueNotFound):
        stub_broker.enqueue(Mock(queue_name="idontexist"))


def test_stub_broker_raises_queue_error_when_joining_on_undeclared_queues(stub_broker):
    # Given that I have a stub broker
    # If I attempt to join on a queue that wasn't declared
    # I expect a QueueNotFound error to be raised
    with pytest.raises(QueueNotFound):
        stub_broker.join("idontexist")


def test_stub_broker_can_be_flushed(stub_broker):
    # Given that I have an actor
    @dramatiq.actor
    def do_work():
        pass

    # When I send that actor a message
    do_work.send()

    # And when there is already a message on that actor's dead-letter queue
    stub_broker.dead_letters_by_queue[do_work.queue_name].append("dead letter")

    # Then its queue should contain the right number of messages
    assert stub_broker.queues[do_work.queue_name].qsize() == 1
    assert len(stub_broker.dead_letters) == 1

    # When I flush all of the queues
    stub_broker.flush_all()

    # Then the queue should be empty
    assert stub_broker.queues[do_work.queue_name].qsize() == 0
    # and it should contain no in-progress tasks
    assert stub_broker.queues[do_work.queue_name].unfinished_tasks == 0
    # and the dead-letter queue should be empty
    assert len(stub_broker.dead_letters) == 0


def test_stub_broker_can_join_with_timeout(stub_broker, stub_worker):
    # Given that I have an actor that takes a long time to run
    @dramatiq.actor
    def do_work():
        time.sleep(1)

    # When I send that actor a message
    do_work.send()

    # And join on its queue with a timeout
    # Then I expect a QueueJoinTimeout to be raised
    with pytest.raises(QueueJoinTimeout):
        stub_broker.join(do_work.queue_name, timeout=500)


def test_stub_broker_join_reraises_actor_exceptions_in_the_joining_current_thread(stub_broker, stub_worker):
    # Given that I have an actor that always fails with a custom exception
    class CustomError(Exception):
        pass

    @dramatiq.actor(max_retries=0)
    def do_work():
        raise CustomError("well, shit")

    # When I send that actor a message
    do_work.send()

    # And join on its queue
    # Then that exception should be raised in my thread
    with pytest.raises(CustomError):
        stub_broker.join(do_work.queue_name, fail_fast=True)
